#!/bin/bash
#
# commitObject -o objectFileName { -s | -f [ -c code ] }
#
#  -o objectFileName  full path to the object to be committed
#  -s                 success
#  -f                 failure
#  -c                 failure code
#
# Valid Failure Codes are:
#
#
#  0110 : A general failure in processing the operation was encountered. [DEFAULT]
#  0112 : No such object instance. One or more of the elements in the Referenced SOP Instance Sequence was not available.
#  0213 : The SCP does not currently have enough resources to store the requested SOP Instance(s).
#  0122 : Storage Commitment has been requested for a SOP Instance with a SOP Class that is not supported by the SCP.
#  0119 : The SOP Class of an element in the Referenced SOP Instance Sequence did not correspond
#         to the SOP class registered for this SOP Instance at the SCP.
#  0131 : The Transaction UID of the Storage Commitment Request is already in use.
#
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Locked by        : $Locker$
# Symbolic name    : $Name$
# State            : $State$
#

. $HOME/etc/setenv.sh

usage_()
{
	[ -n "$1" ] && (echo "$1" ; echo) >&2
	cat << ENDOFUSAGE >&2
  commitObject -o objectFileName { -s | -f <failureMessage> [ -c code ] }

    -o objectFileName  full path to the object to be committed
    -s                 success
    -f                 failure with non empty failure message
    -c                 failure code

 Valid Failure Codes are:

  0110 : A general failure in processing the operation was encountered. [DEFAULT]
  0112 : No such object instance. One or more of the elements in the Referenced SOP Instance Sequence was not available.
  0213 : The SCP does not currently have enough resources to store the requested SOP Instance(s).
  0122 : Storage Commitment has been requested for a SOP Instance with a SOP Class that is not supported by the SCP.
  0119 : The SOP Class of an element in the Referenced SOP Instance Sequence did not correspond
         to the SOP class registered for this SOP Instance at the SCP.
  0131 : The Transaction UID of the Storage Commitment Request is already in use.

ENDOFUSAGE
}


Object=""
Success=""
Failure=""
Code="0110"

while getopts 'o:sfc:' OPT; do
	case $OPT in
		o) Object="$OPTARG" ;;
		s) Success="true" ;;
		f) Failure="true" ;;
		c) Code="$OPTARG" ;;
		*) usage_ "Invalid Parameter: $OPT" ; exit 1 ;;
	esac
done


if [ -z "$Object" -o ! -f "$Object" ] ; then
	usage_ "Invalid file name or file does not exist: $Object"
	exit 1
fi

if [ -z "$Success" -a -z "$Failure" ] ; then
	usage_ "Exactly one of the -s and -f parameters should be specified"
	exit 1
fi

if [ -n "$Success" -a -n "$Failure" ] ; then
	usage_ "Exactly one of the -s and -f parameters should be specified"
	exit 1
fi


ObjectUID=`$DCMTK_HOME/bin/dcmdump +P "0008,0018" +p $Object | grep "^(0008,0018) " | sed 's/^.*\[\(.*\)].*$/\1/'`
if [ $? -ne 0 -o -z "$ObjectUID" ] ; then
	echo "Non DICOM file or file does not contain SOP Instance UID: $Object" >&2
	exit 1
fi

validCode=""
for c in 0110 0112 0213 0122 0119 0131 ; do
	if [ "$Code" = "$c" ] ; then
		validCode="$Code"
		break;
	fi
done

[ -z "$validCode" ] && validCode="0110"

if [ -n "$Success" ] ; then
	echo "success" > $COMMIT_VAR/flags/$ObjectUID
else
	echo "failure:$validCode" > $COMMIT_VAR/flags/$ObjectUID
fi

