/*
 * Component:  dcviewer
 *
 * Purpose:
 *     Extracting raw images/documents from DICOM objects
 *
 */
#ifndef __HERELO_PROCESS_H_INCLUDED__
#define __HERELO_PROCESS_H_INCLUDED__

#include "imgproc/imageprocess.h"
#include "imgproc/rawimage.h"
#ifndef IMGPROC_NO_DCMTK
#include "imgproc/encapsulateddocument.h"
#include "imgproc/pbrcontent.h"
#include "utils/codecregistry.h"
#endif //IMGPROC_NO_DCMTK

class CHereloRawImage;
class CHereloEncapsulatedDocument;

typedef struct
{
	int			m_nSequenceOfUltrasoundRegions;
	uint32_t	m_uMinX, m_uMinY, m_uMaxX, m_uMaxY;
	int32_t		m_nRefPixX, m_nRefPixY;
	uint16_t	m_uUnitX, m_uUnitY;
	double		m_dRefPixX, m_dRefPixY, m_dUnitX, m_dUnitY;
} UltrasoundRegionInfo;

/**
 * Memory manager for extracted images/documents created by herelo
 */
class CHereloMemory : public CImageMemory {
	private:

		/** Allocated memory block */
		char * m_block;

		/** Size of the allocated memory block */
		unsigned int m_length;

		/** Indicates whether the block is in use or not */
		bool m_isInUse;

	public:

		/**
		 * Constructor for creating a new herelo memory object
		 */
		CHereloMemory();

		/**
		 * Destructor for destroying this herelo memory object
		 */
		virtual ~CHereloMemory();

		/**
		 * Allocates memory block for an extracted image/document
		 *
		 * @param length   Number of bytes to be allocated
		 * @param index    Image (frame) index within the DICOM object
		 *
		 * @return a pointer to the allocated memory block or NULL
		 */
		virtual char * allocate( unsigned int length, int index );

		/**
		 * Releases memory of an extracted image/document
		 *
		 * @param block    Memory block to be released
		 * @param length   Number of allocated bytes in the block
		 * @param index    Image (frame) index within the DICOM object
		 */
		virtual void release( char * block, unsigned int length, int index );

		/**
		 * Chunks allocated memory block of an extracted image/document.
		 * New length should be less or equal than the original length.
		 * The chunked memory block might differ from the original one.
		 *
		 * @param block       Memory block to be chunked
		 * @param oldLength   Number of allocated bytes in the block
		 * @param newLength   Number of bytes used in the block (chunked length)
		 * @param index       Image (frame) index within the DICOM object
		 *
		 * @return a pointer to chunked memory block or NULL
		 */
		virtual char * chunk( char * block, unsigned int oldLength, unsigned int newLength, int index );
};


class CFrameInfo {
	private:

		/** Flag indicating which fields are provided for this frame */
		unsigned long m_flag;

		// ------
		// -- Values already handled by CImageProcess and stored on root level
		// ------

		// The viewer stores all w/l settigns, we store only the first one
		//		int             m_nWindowCenterCnt;
		//		int             m_nWindowWidthCnt;
		//		int             m_nWindowCenterWidthExplanationCnt;
		//		CList<CWEInfo*, CWEInfo*>* m_pWindowCWEInfo;

		/** Window Center */
		double m_windowCenter;

		/** Window Width */
		double m_windowWidth;


		/** Rescale Intercept */
		double m_rescaleIntercept;

		/** Rescale Slope */
		double m_rescaleSlope;

		/** Pixel Spacing X */
		double m_pixelSpacingX;

		/** Pixel Spacing Y */
		double m_pixelSpacingY;

		/** Imager Pixel Spacing X */
		double m_imagerPixelSpacingX;

		/** Imager Pixel Spacing Y */
		double m_imagerPixelSpacingY;

		/** Nominal Scanned Pixel Spacing X */
		double m_nominalScannedPixelSpacingX;

		/** Nominal Scanned Pixel Spacing Y */
		double m_nominalScannedPixelSpacingY;

		// -----
		// -- Values not parsed by CImageProcess
		// -----


		/** Slice thickness */
		double m_sliceThickness;

		/** Spacing Between Slices */
		double m_spacingBetweenSlices;


		// Viewer parses patient orientation data, we only want to add it to the desc file
		// PatientOrientationInfo m_oPatientOrientation;

		// OFString m_frameOfReferenceUID; // root level only

		OFString m_imagePositionPatient;
		OFString m_imageOrientationPatient;
		OFString m_patientOrientation;


		// -----
		// The fields below are processed by the viewer at frame level, but we ignore it for now
		//
		//		LUTInfo* m_pModalityLUTInfo;
		//		CList<LUTInfo*, LUTInfo*>* m_pVOILUTInfo;
		//
		//		double  m_dImagerPixelSpacingX;
		//		double  m_dImagerPixelSpacingY;
		//		double  m_dNominalScannedPixelSpacingX;
		//		double  m_dNominalScannedPixelSpacingY;
		//		double  m_dEstimatedRadiographicMagnificationFactor;

	public:
		CFrameInfo();
		~CFrameInfo();


		inline const OFString& getImageOrientationPatient() const {
			return m_imageOrientationPatient;
		}

		inline void setImageOrientationPatient( const OFString& imageOrientationPatient ) {
			m_flag |= FlagImageOrientPatient;
			m_imageOrientationPatient = imageOrientationPatient;
		}

		inline const OFString& getImagePositionPatient() const {
			return m_imagePositionPatient;
		}

		inline void setImagePositionPatient( const OFString& imagePositionPatient ) {
			m_flag |= FlagImagePosPatient;
			m_imagePositionPatient = imagePositionPatient;
		}

		inline const OFString& getPatientOrientation() const {
			return m_patientOrientation;
		}

		inline void setPatientOrientation( const OFString& patientOrientation ) {
			m_flag |= FlagPatientOrient;
			m_patientOrientation = patientOrientation;
		}

		inline double getPixelSpacingX() const {
			return m_pixelSpacingX;
		}

		inline double getPixelSpacingY() const {
			return m_pixelSpacingY;
		}

		inline void setPixelSpacing( double x, double y ) {
			m_flag |= FlagPixelSpacing;
			m_pixelSpacingX = x;
			m_pixelSpacingY = y;
		}

		inline double getImagerPixelSpacingX() const {
			return m_imagerPixelSpacingX;
		}

		inline double getImagerPixelSpacingY() const {
			return m_imagerPixelSpacingY;
		}

		inline void setImagerPixelSpacing( double x, double y ) {
			m_imagerPixelSpacingX = x;
			m_imagerPixelSpacingY = y;
		}

		inline double getNominalScannedPixelSpacingX() const {
			return m_nominalScannedPixelSpacingX;
		}

		inline double getNominalScannedPixelSpacingY() const {
			return m_nominalScannedPixelSpacingY;
		}

		inline void setNominalScannedPixelSpacing( double x, double y ) {
			m_nominalScannedPixelSpacingX = x;
			m_nominalScannedPixelSpacingY = y;
		}

		inline double getRescaleIntercept() const {
			return m_rescaleIntercept;
		}

		inline void setRescaleIntercept( double rescaleIntercept ) {
			m_flag |= FlagRescaleIntercept;
			m_rescaleIntercept = rescaleIntercept;
		}

		inline double getRescaleSlope() const {
			return m_rescaleSlope;
		}

		inline void setRescaleSlope( double rescaleSlope ) {
			m_flag |= FlagRescaleSlope;
			m_rescaleSlope = rescaleSlope;
		}

		inline double getSliceThickness() const {
			return m_sliceThickness;
		}

		inline void setSliceThickness( double sliceThickness ) {
			m_flag |= FlagSliceThickness;
			m_sliceThickness = sliceThickness;
		}

		inline double getSpacingBetweenSlices() const {
			return m_spacingBetweenSlices;
		}

		inline void setSpacingBetweenSlices( double spacingBetweenSlices ) {
			m_flag |= FlagSliceSpacing;
			m_spacingBetweenSlices = spacingBetweenSlices;
		}


		inline double getWindowCenter() const {
			return m_windowCenter;
		}

		inline void setWindowCenter( double windowCenter ) {
			m_flag |= FlagWindowCenter;
			m_windowCenter = windowCenter;
		}

		inline double getWindowWidth() const {
			return m_windowWidth;
		}

		inline void setWindowWidth( double windowWidth ) {
			m_flag |= FlagWindowWidth;
			m_windowWidth = windowWidth;
		}

		inline bool isSet( unsigned long flag ) const {
			return ( m_flag & flag ) == flag;
		}

		// Static flag values

		static unsigned long FlagWindowCenter;
		static unsigned long FlagWindowWidth;
		static unsigned long FlagRescaleSlope;
		static unsigned long FlagRescaleIntercept;
		static unsigned long FlagPixelSpacing;
		static unsigned long FlagSliceThickness;
		static unsigned long FlagSliceSpacing;
		static unsigned long FlagPatientOrient;
		static unsigned long FlagImageOrientPatient;
		static unsigned long FlagImagePosPatient;
};



/**
 * DICOM object processing.
 * Processes dicom dataset and extracts images/document stored in the object.
 */
class CHereloProcess : public CImageProcess {
	private:

#ifndef IMGPROC_NO_DCMTK
		/** Registering supported codecs */
		CodecRegistry * m_codecs;
#endif //IMGPROC_NO_DCMTK

		/** Output stream for the object descriptor file to hold information an extracted pixel data */
		FILE * m_descFile;

		/** Prefix for saving the processed (herelt) DICOM object (and files intended to be put to the processed repository) */
		OFString m_outputPrefix;

		/** Prefix for saving the extracted pixel data (files intended to be put to the cache repository) */
		OFString m_cachePrefix;

		/** Don't create cached */
		bool m_noCached;

		/** Number of supported private tags */
		int m_lengthIncludeArray;

#ifndef IMGPROC_NO_DCMTK
		/** Array of supported private tags (not to be removed from the processed DICOM object) */
		DcmTag * m_includeArray;
#endif //IMGPROC_NO_DCMTK

		/** Indicates whether bitmap optimization is to be performed or not */
		bool m_needOptimize;

		// Herelo specific tags

		/** Frame of reference UID */
		OFString m_frameOfReferenceUID;

		/** Common frame data */
		CFrameInfo m_commonFrame;

		/** Per frame data */
		CFrameInfo * m_frames;

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Reads configuration file storing supported private tags
		 *
		 * @param filename   path to the file storing supported private tags
		 */
		void readIncludeConfig( const char * filename );

		/**
		 * Register compression codecs (RLE, JPEG, JPEG2000)
		 */
		void registerCodecs();

		/**
		 * Deregister compression codecs (RLE, JPEG, JPEG2000)
		 */
		void deregisterCodecs();
#endif //IMGPROC_NO_DCMTK

		/**
		 * Write extracted images, generated thumbnails to files
		 *
		 * @param image   Holds pixel data of the extracted images
		 *
		 * @return  OFTrue if and only if the images could be successfully written
		 */
		OFBool writeImageFiles( CHereloRawImage * image );

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Write extracted document to file
		 *
		 * @param document   Holds extracted document
		 *
		 * @return  OFTrue if and only if the document could be successfully written
		 */
		OFBool writeDocumentFiles( CHereloEncapsulatedDocument * document );

		/**
		 * Write obsolescence information to file
		 *
		 * @param pbrContent   Holds obsolescence information extracted from PbR files
		 *
		 * @return  OFTrue if and only if the obsolescence information could be successfully written
		 */
		OFBool writeObsInfo( CPbRContent * pbrContent );

		/**
		 * Write study fingerprint to file
		 *
		 * @param pbrContent   Holds study fingerprint extracted from PbR files
		 *
		 * @return  OFTrue if and only if the fingerprint could be successfully written
		 */
		OFBool writeFingerPrint( CPbRContent * pbrContent );
#endif //IMGPROC_NO_DCMTK

	protected:

		/** Compression flags (default: 0)
		 *  1 - create cw4 thumbs (in processed)
		 *  2 - create cw4 images (in cached)
		 *  4 - do not cw1 thumbs */
		int m_compressionFlags;

		/** cw4 thumb rate: target fraction of uncompressed image size (default: 0.01) */
		double m_cmpThumbrate;

		/**
		 * Creates a new CHereloRawImage object.
		 *
		 * @return pointer to the newly created CHereloRawImage object
		 */
		virtual CRawImage * createRawImage();

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Creates a new CHereloEncapsulatedDocument object.
		 *
		 * @param type     Document type
		 * @param length   Length of the encapsulated document in bytes
		 *
		 * @return pointer to the newly created CHereloEncapsulatedDocument object
		 */
		virtual CEncapsulatedDocument * createEncapsulatedDocument( DOCUMENT_TYPE type, Uint32 length );
#endif //IMGPROC_NO_DCMTK

		/**
		 * Post processing extracted data
		 * (eg histogram analyzis, planar, color conversions, ... for image frames)
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool postProcess( CExtractedData * data );

		/**
		 * Initialization for DICOM dataset processing.
		 * Herelo registers codecs here
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool initProcess();

		/**
		 * Finalization of DICOM dataset processing.
		 * Reference and calculated values will be filled in the DICOM dataset
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool finalizeProcess();

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Decides if the Private DICOM Tag should remain in the dataset or can be removed.
		 *
		 * @param tag   Private DICOM Tag to be checked
		 *
		 * @return OFTrue if the tag should be kept in the DICOM dataset
		 */
		virtual OFBool isIncluded( const DcmTag & tag );
#endif //IMGPROC_NO_DCMTK

	public:

		void setCompressionOptions( int compressionFlags ) {
			m_compressionFlags = compressionFlags;
		}

		void setThumbCmpRate( double cmpThumbRate ) {
			m_cmpThumbrate = cmpThumbRate;
		}

		void dontCreateCached() {
			m_noCached = true;
		}

		void initFunction2() /* override*/ ;

		/**
		 * Sets flag indicating whether bitmap should be optimized or not
		 *
		 * @param needOptimize  true indicates bitmap needs to be optimized
		 */
		inline void setNeedOpt( bool needOptimize ) {
			m_needOptimize = needOptimize;
		}

		/**
		 * @return  true if and only if the flag for bitmap optimization is set.
		 */
		inline bool getNeedOpt() {
			return m_needOptimize;
		}

		/**
		 * Constructor for creating a new CHereloProcess object
		 *
		 * @param dataset         DICOM dataset to be processed
		 * @param verbose         OFTrue indicates verbose mode, the dataset content will be dumped
		 * @param dumponly        OFTrue indicates to dump only, do not process the dataset
		 * @param escapeNewLines  OFTrue indicates to escape new lines in the dumps
		 * @param outputPrefix    Path prefix to save processed object (processed repository).
		 * @param cachePrefix     Path prefix to save image(s) (cache repository).
		 * @param specificFrame   Number of specific frame to process, if non-zero,
		 *                        only image data for the specific frame will be processed and written,
		 *                        descriptor and processed dicom WILL NOT be written,
		 */
#ifndef IMGPROC_NO_DCMTK
		CHereloProcess( DcmDataset * dataset, OFBool verbose, OFBool dumpOnly, OFBool escapeNewLines, const char * outputPrefix, const char * cachePrefix, int specificFrame);
#else
		CHereloProcess( DcmDataset * dataset, const char * outputPrefix, const char * cachePrefix, int specificFrame);
#endif //IMGPROC_NO_DCMTK

		/**
		 * Destructor for destroying this CHereloProcess object
		 */
		virtual ~CHereloProcess();


		/**
		 * Return frame specific information for the requested frame
		 *
		 * @param frameNdx   frame index (0 based)
		 *
		 * @return the frame specific information, or the common frame information for unrecognized frames
		 */
		inline const CFrameInfo & getFrameInfo( int frameNdx ) const {
			return ( m_frames == NULL || frameNdx >= m_frameNumber ) ? m_commonFrame : m_frames[ frameNdx ];
		}

		/**
		 * Return common (base) frame information
		 *
		 * @return the common (base) frame information
		 */
		inline const CFrameInfo & commonFrameInfo() const {
			return m_commonFrame;
		}

		/**
		 * @return the Frame of Reference UID
		 */
		inline const OFString & getFrameOfReferenceUID() const {
			return m_frameOfReferenceUID;
		}

#ifndef IMGPROC_NO_DCMTK
	protected:


		static void fHereloTag2DCM_SharedFunctionalGroupsSequence( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_PerFrameFunctionalGroupsSequence( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_PatientOrientation( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_ImagePositionPatient( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_ImageOrientationPatient( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_FrameOfReferenceUID( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_SliceThickness( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_SpacingBetweenSlices( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_DetectorInformationSequence( IDicomTagParser2 * obj );


		// Already defined in CImageProcess - but to put the extracted data to frameInfo
		static bool readUltrasoundRegionData(DcmElement* pElement, UltrasoundRegionInfo* pUltrasoundRegionInfo);
		static void fHereloTag2DCM_SequenceOfUltrasoundRegions( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_PixelSpacing( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_ImagerPixelSpacing( IDicomTagParser2 *obj );
		static void fHereloTag2DCM_NominalScannedPixelSpacing( IDicomTagParser2 *obj );
		static void fHereloTag2DCM_WindowCenter( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_WindowWidth( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_RescaleIntercept( IDicomTagParser2 * obj );
		static void fHereloTag2DCM_RescaleSlope( IDicomTagParser2 * obj );

		// fTag2DCM_WindowCenter
		// fTag2DCM_WindowWidth
		// fTag2DCM_RescaleIntercept
		// fTag2DCM_RescaleSlope

		// Frame number is already read by CImageProcess::fTag2DCM_NumberOfFrames,
		// but we need to create the frame array for multiframe objects
		static void fHereloTag2DCM_NumberOfFrames( IDicomTagParser2 * obj );
#endif //IMGPROC_NO_DCMTK

};

/**
 * Fills raw image bitmap, calculates best window, best level,
 * makes planar and colorspace transformations.
 */
class CHereloRawImage : public CRawImage {
	protected:

		/**
		 * Converts bitmap stored on 2 bytes per smaple to 1 byte per sample if it's possible
		 */
		virtual void optimizeBitmap();

	public:

		/**
		 * Constructor for creating new CHereloRawImage object.
		 *
		 * @param imageProcess   Image Process object holding original DICOM data
		 * @param index          Image (frame) index within the DICOM object
		 */
		CHereloRawImage( CImageProcess * imageProcess, int index );

		/**
		 * Destructor for destroying this CHereloRawImage object.
		 */
		virtual ~CHereloRawImage();

		/**
		 * Initializes CHereloRawImage object for processing.
		 * Gets frame specific image data (W/L, pixel spacing, ...)
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool init() /* override */;


		/**
		 * Prints pixel data information (size, depth, window/level, ...) to image descriptor file stream
		 *
		 * @param fout   stream to image descriptor file
		 *
		 * @return OFTrue if and only if pixel information was successfully written to the stream
		 */
		OFBool printDesc( FILE * fout );

		/**
		 * Writes raw pixel data to file stream
		 *
		 * @param fImage  file stream the raw pixel data to be written to
		 *
		 * @return OFTrue if and only if pixel data was successfully written to the file stream
		 */
		OFBool writeRawFile( FILE * fImage );

		/**
		 * Creates 128x128 pixel thumbnail from image
		 *
		 * @param fThumb   file stream for the thumbnail to be created
		 * @param fCThumb  file stream for the wavelet thumbnail image to be created
		 *
		 * @return  OFTrue if and only if the thumbnail was successfully created and written to the file stream
		 */
		OFBool createAndWriteThumbnail( FILE * fThumb, FILE * fCThumb );

		/**
		 * Creates and writes the full compressed stream for the image,
		 * and the stating chunk of the stream beginning as the thumbnail
		 *
		 * @param fCmpImage    file stream the full stream is to be written to (NULL to skip)
		 * @param fCmpThumb    file stream the beginning of the stream to be written to (NULL to skip)
		 * @param dThumbRatio  fraction of the full compressed stream to be written to fCmpThumb
		 *
		 * @return  OFTrue if and only if the files were successfully created and written
		 */
		OFBool createAndWriteCompressedStreamElements( FILE * fCmpImage, FILE * fCmpIQ, FILE * fCmpThumb, double dThumbRatio );

	private:
		OFBool writeEPIDHeader( CHereloProcess * proc, FILE * f );

};

#ifndef IMGPROC_NO_DCMTK

/**
 * Extracting document from Encapsulated Document DICOM objects.
 */
class CHereloEncapsulatedDocument : public CEncapsulatedDocument {
	public:
		/**
		 * Constructor for creating a new Encapsulated Document.
		 *
		 * @param imageProcess   Image Process object holding original DICOM data
		 * @param type           Document type (Currently the only supported type is PDF)
		 * @param length         Length of the Document
		 */
		CHereloEncapsulatedDocument( CImageProcess * imageProcess, DOCUMENT_TYPE type, Uint32 length );

		/**
		 * Destructor for destroying this encapsulated document
		 */
		virtual ~CHereloEncapsulatedDocument();

		/**
		 * Writes extracted document to file stream
		 *
		 * @param fDoc  file stream the document to be written to
		 *
		 * @return OFTrue if and only if the document was successfully written to the file stream
		 */
		OFBool writeDocumentFile( FILE * fDoc );
};

#endif //IMGPROC_NO_DCMTK

#endif // __HERELO_PROCESS_H_INCLUDED__
