/**
 * Component: imgproc
 *
 * Description:
 *    Class for extracting obsolescence information from PbR objects
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */

#ifndef CPBRCONTENT_H_
#define CPBRCONTENT_H_

#include "dcmtk/ofstd/ofstd.h"
#include "dcmtk/ofstd/ofdate.h"
#include "dcmtk/ofstd/oftime.h"

#include "imgproc/extracteddata.h"

class CImageProcess;

/**
 * Obsolescence modes
 */
typedef enum {
	OBSCOLESCENCE_NONE,  // Not obsolete
	OBSCOLESCENCE_STUDY, // Whole study is obsolete
	OBSCOLESCENCE_PARTIAL // Part of the study is obsolete
} OBSOLESCENCE_MODE;

/**
 * Information on a single invalidated series
 */
class CInvalidatedSeries {
private:
	/** Invalidate Date */
	OFDate m_invDate;

	/** Invalidate Time */
	OFTime m_invTime;

	/** Series Instance UID of the invalidated series */
	OFString m_uid;

public:

	/**
	 * Constructor to create a new invalidate series instance
	 *
	 * @param uid       Series Instance UID of the invalidated series
	 * @param invDate   Invalidate Date
	 * @param invTime   Invalidate Time
	 */
	CInvalidatedSeries(const char *uid, const OFDate &invDate,
			const OFTime &invTime) {
		m_uid = uid;
		m_invDate = invDate;
		m_invTime = invTime;
	}

	/**
	 * @return Series Instance UID of the invalidated series
	 */
	inline const OFString& getUid() const {
		return m_uid;
	}

	/**
	 * @return Invalidate Date
	 */
	inline const OFDate& getInvalidationDate() const {
		return m_invDate;
	}

	/**
	 * @return Invalidate Time
	 */
	inline const OFTime& getInvalidationTime() const {
		return m_invTime;
	}
};

/**
 * Information on a single invalidated object
 */
class CInvalidatedObject {
private:
	/** Invalidate Date */
	OFDate m_invDate;

	/** Invalidate Time */
	OFTime m_invTime;

	/** Filename of the invalidated object */
	OFString m_fileName;

public:

	/**
	 * Constructor to create a new invalidate object instance
	 *
	 * @param fileName  Filename of the invalidated object
	 * @param invDate   Invalidate Date
	 * @param invTime   Invalidate Time
	 */
	CInvalidatedObject(OFString fileName, const OFDate &invDate,
			const OFTime &invTime) {
		m_fileName = fileName;
		m_invDate = invDate;
		m_invTime = invTime;
	}

	/**
	 * @return Filename of the invalidated object
	 */
	inline const OFString& getFileName() const {
		return m_fileName;
	}

	/**
	 * @return Invalidate Date
	 */
	inline const OFDate& getInvalidationDate() const {
		return m_invDate;
	}

	/**
	 * @return Invalidate Time
	 */
	inline const OFTime& getInvalidationTime() const {
		return m_invTime;
	}
};

/**
 * Class for extracting obsolescence information from PbR objects
 */
class CPbRContent: public CExtractedData {
private:

	/** Obsolescence mode */
	OBSOLESCENCE_MODE m_obsMode;

	/** Invalidate Date */
	OFDate m_invDate;

	/** Invalidate Time */
	OFTime m_invTime;

	/** List of obsolete object files */
	OFList<OFString> m_obsFiles;

	/** Stored fingerprint with the state of study at report time */
	OFString m_fingerprint;

	/** Invalidated series */
	OFList<CInvalidatedSeries*> m_invSeries;

	/** Invalidated objects */
	OFList<CInvalidatedObject*> m_invObjects;

public:

	/**
	 * Constructor for creating new CPbRContent object.
	 *
	 * @param imageProcess   Image Process object holding original DICOM data
	 */
	CPbRContent(CImageProcess *imageProcess);

	/**
	 * Destructor for destroying CPbRContent object.
	 */
	virtual ~CPbRContent();

	/**
	 * Initializes CPbRContent object for processing.
	 *
	 * @return OFTrue on success; OFFalse on error
	 */
	virtual OFBool init();

	/**
	 * Post processes extracted data.
	 *
	 * @return 0 on success; >0 on error
	 */
	virtual int postProcess();

	/**
	 * @return OFTrue if either the whole study is obsolete or there are obsolete/invalidated objects in the study
	 */
	inline OFBool hasObsolescenceInfo() const {
		return m_obsMode != OBSCOLESCENCE_NONE;
	}

	/**
	 * @return  OFTrue if the entire study is obsolete
	 */
	inline OFBool isWholeStudyObsolete() const {
		return m_obsMode == OBSCOLESCENCE_STUDY;
	}

	/**
	 * @return  OFTrue if there are invalidated objects in the study
	 */
	inline OFBool isInvalidated() const {
		return m_invDate.isValid();
	}

	/**
	 * @return Invalidation Date
	 */
	inline const OFDate& getInvalidateDate() const {
		return m_invDate;
	}

	/**
	 * @return Invalidation Time
	 */
	inline const OFTime& getInvalidateTime() const {
		return m_invTime;
	}

	/**
	 * @return  OFTrue if there are obsolete objects in the study
	 */
	inline OFBool hasObsoleteObjects() const {
		return !m_obsFiles.empty();
	}

	/**
	 * Returns the list of the obsolete objects. the list will contain the short file names of the obsolete objects.
	 * @return list of the obsolete objects
	 */
	inline const OFList<OFString>& getObsoleteObjects() const {
		return m_obsFiles;
	}

	/**
	 * @return  OFTrue if there are invalidated series in the study
	 */
	inline OFBool hasInvalidatedSeries() const {
		return !m_invSeries.empty();
	}

	/**
	 * Returns the list of the invalidated series. the list will contain the UID and timestamp of the invalidated series.
	 * @return list of the invalidated series
	 */
	inline const OFList<CInvalidatedSeries*>& getInvalidatedSeries() const {
		return m_invSeries;
	}

	/**
	 * @return  OFTrue if there are invalidated objects in the study
	 */
	inline OFBool hasInvalidatedObject() const {
		return !m_invObjects.empty();
	}

	/**
	 * Returns the list of the invalidated objects. the list will contain the filename and timestamp of the invalidated objects.
	 * @return list of the invalidated objects
	 */
	inline const OFList<CInvalidatedObject*>& getInvalidatedObjects() const {
		return m_invObjects;
	}

	/**
	 * @return Stored fingerprint describing the state of study at report time
	 */
	inline const OFString& getFingerPrint() const {
		return m_fingerprint;
	}
};

#endif /* CPBRCONTENT_H_ */
