/*
 * Component:  common
 *
 * Purpose:
 *	    Header file for DICOM image object / raw image processing
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __RAW_IMAGE_H_INCLUDED__
#define __RAW_IMAGE_H_INCLUDED__

#ifndef IMGPROC_NO_DCMTK
#include "dcmtk/ofstd/oftypes.h"
#include "dcmtk/ofstd/ofstring.h"
#else
#include "imgproc/nodcmtk.h"
#endif //IMGPROC_NO_DCMTK

#include "imgproc/dicomdefine.h"

#include "imgproc/extracteddata.h"

/**
 * Fills raw image bitmap, calculates best window, best level,
 * makes planar and colorspace transformations.
 */
class CRawImage : public CExtractedData {
	protected:

		/** Current byte in m_contentData */
		char * m_current;

		// Min/max pixel values

		/** type based on really used bits (8/16) */
		int m_type;

		/** Bytes per pixel (1 - Monochrome byte / 2 - Monochrome word / 3 - color byte) */
		int m_bytesPerPixel;

		/** Rescale Intercept, will be re-calculated */
		double m_rescaleIntercept;

		/** Rescale Slope, will be re-calculated */
		double m_rescaleSlope;

		/** Window Center, will be re-calculated */
		double m_windowCenter;

		/** Window Width, will be re-calculated */
		double m_windowWidth;

		/** Indicates whether color image pixels are stored by plane (true) or by pixel (false) */
		bool m_isPlane;

		/** Samples Per Pixel - might be changed if a 16 bits image used only 8 bits */
		Uint16 m_samplesPerPixel;

		/** Image Type in human readable format (will be printed to desc file) */
		OFString m_imageTypeName;

		/** Number of bits are really used */
		int m_bitsUsed;

		/** Blank image is solid black one */
		bool m_isBlank;

		/**
		 * Shift pixel values to be zero based
		 *
		 * @param minValue    smallest pixel value in the raw image, this value will be the 0
		 */
		void shiftImageToOrigo( Sint32 minValue );

		/**
		 * Converts RawPerPlane into RawPerPixel (RRRRGGGGBBBB -> RGBRGBRGBRGB)
		 *
		 * @param data     pixel data to be converted
		 * @param length   length of pixel data
		 */
		void planarToInterleaved3( char * data, int length );

		/**
		 * Converts color images into RGB.
		 * Working directly with m_rawImage
		 */
		void convertColor2RGB();

		/**
		 * converts bitmap stored on 2 bytes per smaple to 1 byte per sample if it's possible
		 */
		virtual void optimizeBitmap();

	public:

		/**
		 * Constructor for creating new CRawImage object.
		 *
		 * @param imageProcess   Image Process object holding original DICOM data
		 * @param index          Image (frame) index within the DICOM object
		 */
		CRawImage( CImageProcess * imageProcess, int index );

		/**
		 * Destructor for destroying CRawImage object.
		 */
		virtual ~CRawImage();

		/**
		 * Initializes CRawImage object for processing.
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool init();

		/**
		 * Adds word sample to the raw image data. Will be used for Palette Color images
		 *
		 * @param sample   Word sample to be added
		 */
		void addSample( Uint16 sample );

		/**
		 * Adds byte samples to the raw image data.
		 *
		 * @param buffer   Byte samples to be added
		 * @param length   Count of samples to be added
		 */
		void addSamples( Uint8 * buffer, Uint32 length );

		/**
		 * Adds word samples to the raw image data.
		 *
		 * @param buffer   Word samples to be added
		 * @param length   Count of samples to be added
		 */
		void addSamples( Uint16 * buffer, Uint32 length );


		/**
		 * Adds byte samples from word samples to the raw image data.
		 * Will use the low byte of the word values.
		 *
		 * @param buffer   Word samples to be added
		 * @param length   Count of samples to be added
		 */
		void addByteSamples( Uint16 * buffer, Uint32 length );



		/**
		 * Post processes raw image data including histogram analyzis, colorspace conversion,
		 * planar conversion, ...
		 *
		 * @return 0 on success; >0 on error
		 */
		virtual int postProcess();

		/**
		 * Set RawImage to blank.
		 *
		 */
		void setBlank();

		/**
		 * Is the RawImage blank.
		 *
		 * @return blank state;
		 */
		bool isBlank();
};

#endif // __RAW_IMAGE_H_INCLUDED__
