#pragma once
#include "imutils/GCPConnector.h"
#include "imutils/IMProperties.h"
#include <curl/curl.h>
#include <dcmtk/ofstd/ofstring.h>
#include <google/cloud/credentials.h>
#include <nlohmann/json.hpp>

namespace erad::gcp {

class CloudServiceConnector : public GCPConnector {
	public:
		struct UploadDicomFileData {
				OFString filepath;
				// Optional name for the target file (If a change is needed)
				OFString targetFileName;
				OFString studyInstanceUID;
				OFString seriesInstanceUID;
				OFString instanceUID;
				OFBool forcefulUpload;
		};

		/**
		 * Constructor for initialization GCP Network for cloud service
		 *
		 * @param deviceID         ID of the device
		 * @param dicomStorePath   GCP dicomStore Path
		 *
		 * @throws IMException if an error occures during processing the request
		 */
		CloudServiceConnector( const OFString& deviceID, const OFString& dicomStorePath );

		/**
		 * Retrieve the dicom file.
		 *
		 * @param studyId   study instance uid
		 * @param seriesId   series instance uid
		 * @param instanceId   instance uid
		 * @param tempDicomFilePath   temporary dicom file path
		 *
		 * @return true if success
		 */
		OFBool retrieveDicomFile( const OFString& studyId, const OFString& seriesId, const OFString& instanceId,
				const OFString& tempDicomFilePath );

		/**
		 * Issues an upload to GCP Healthcare API.
		 *
		 * @param data   UploadDicomFile attributes
		 *
		 * @return EC_Normal if the object was successfully sent or an error code
		 */
		OFBool uploadDicomFileHealthcareAPI( const UploadDicomFileData& data );

		/**
		 * Issues an upload to GCP bucket.
		 *
		 * @param filepath   the DICOM object path
		 * @param parentFolder   the parent folder path
		 * @param filename   the file name
		 *
		 * @return EC_Normal if the object was successfully sent or an error code
		 */
		OFBool uploadFileToGCS( const OFString& filepath, const OFString& parentFolder, const OFString& filename );

		/**
		 * Destructor for destroying the CloudServiceConnector object
		 */
		~CloudServiceConnector();
	
	private:
		struct ResponseData {
				long response_code;
				std::string responseText;
				bool moveToGCS;
				std::string dicomFilePath;
				// Optional name for the target file (If a change is needed)
				OFString targetFileName;
				OFString studyInstanceUID;
				OFString seriesInstanceUID;
				OFString instanceUID;
				bool forcefulUpload;
		};
		/** constants for gcp healthcare request timeout values*/
		const long REQUEST_TIMEOUT = 600; // 10 minutes in seconds
		const long CONNECTION_TIMEOUT = 30; // 30 seconds

		/** Dicom store path for the gcp healthcare requests*/
		std::string m_dicomStorePath;
		/** Device id for the gcp healthcare or gcs device*/
		std::string m_gcpDeviceID;

		/**
		 * Issues an upload to GCP Healthcare API.
		 *
		 * @param studyInstanceUID
		 * @param seriesInstanceUID
		 * @param instanceUID
		 *
		 * @return true delere from the healthcare api
		 */
		OFBool deleteFileFromGCP( const OFString& studyInstanceUID, const OFString& seriesInstanceUID, const OFString& instanceUID );

		/**
		 * Handle the response data from GCP.
		 *
		 * @param data   Response data container object
		 *
		 * @return true if success
		 */
		OFBool handleResponse( ResponseData& data);

		/**
		 * Process the response data from GCP.
		 *
		 * @param data   Response data container object
		 *
		 * @return true if success
		 */
		OFBool processUploadHealthcareApiResponse( ResponseData& data );

		/**
		 * Handle gcp response for upload to bucket and delete pbr 
		 *
		 * @param curl   curl object
		 * @param operation operation on cloud like delete upload etc
		 * @param resourceName   GCP resource name
		 * @return true if success
		 */
		OFBool handleGCPResponseCode( CURL* curl, const std::string& operation, const std::string& resourceName );

		/**
		 * Set up curl as part of network request initialization.
		 *
		 * @param accessToken   access token
		 *
		 * @return true if success
		 */
		virtual OFBool setupCurlForCloudServiceType( const OFString& accessToken ) override;

		/**
		 * Helper function to parse the data retrieved from GCP Healthcare
		 *
		 * @param input   response from gcp
		 *
		 * @return Extracted boundary line
		 */
		OFString extractBeforeContentType( const OFString& input );

		/**
		 * Helper function to createFinalBoundaryString to parse the data retrieved from GCP Healthcare
		 *
		 * @param extracted   response from gcp
		 *
		 * @return Extracted final boundary line
		 */
		OFString createFinalBoundaryString( const OFString& extracted );

		/**
		 * Helper function to remove extraction from input to parse the data retrieved from GCP Healthcare
		 *
		 * @param input   response from gcp
		 * @param extracted   extracted response from gcp
		 *
		 * @return Extracted final boundary line
		 */
		OFString removeExtractedFromInput( const OFString& input, const OFString& extracted );

		/**
		 * function to process retrieved response from GCP Healthcare
		 *
		 * @param response   response from gcp
		 * @param tempDicomFilePath   temporary dicom FilePath
		 *
		 * @return true if sucessfully processed response
		 */
		OFBool processRetrieveDicomResponse( const OFString& response, const OFString& tempDicomFilePath );

		/**
		 * function to set config path
		 * 
		 * @throws IMException if an error occures setting config path
		 *
		 * @return device config path
		 */
		OFString getConfigPath();
};

} // namespace erad::gcp
