/*
 * Interface definition and MySQL implementation to access database connections.
 * Also provides a cache for MySQL prepared statements.
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_DBCONNECTOR_H_
#define IMUTILS_INCLUDE_IMUTILS_DBCONNECTOR_H_

#include "imutils/Sql.h"
#include <string>
#include <functional>
#include <map>
#include <thread>
#include <random>
#include <memory>
#include <imutils/ThreadCache.h>

namespace erad::db {

class RefcountDB;
class StudyDB;
class ObjectDB;
class MiscDB;
class LockDB;
class SessionDB;
class RawQuery;


/* ********************************************
 *
 * Interface
 *
 * ********************************************/


/**
 * Exceptions used by the Database Encapsulation Layer (DBEL)
 */
class DBException : public std::exception {

		/// the error message
		std::string message;

		// error code (for bubbling up underlying sql exception codes)
		unsigned int errorCode;

	public:

		/**
		 * Constructs a new DB Exception having an error message provided by the underlying
		 * database implementation.
		 *
		 * @param message the error message
		 */
		DBException( const std::string& message ) noexcept;

		DBException( const std::string& message, unsigned int errorCode ) noexcept;

		/* Use the default copy and move constructors */

		DBException( const DBException& orig ) noexcept = default;
		DBException( DBException&& orig ) noexcept = default;

		virtual ~DBException();

		/* Use the default copy and move assignments */

		DBException& operator=( const DBException& ) noexcept = default;
		DBException& operator=( DBException&& ) noexcept = default;

		inline unsigned int getErrorCode() const noexcept {
			return errorCode;
		}

		/**
		 * @copydoc std::exception::what()
		 */
		virtual const char* what() const noexcept override {
			return message.c_str();
		}
};


/**
 * Interface to get connections to the medsrv database(s) and tables.
 * Part of the Database Encapsulation Layer (DBEL).
 */
class DBConnector {
	private:
		std::random_device rand_dev;
		std::mt19937 generator;
		std::uniform_int_distribution<int> distr;

	protected:

		/**
		 * Default constructor is only accessible from derived classes.
		 */
		DBConnector() noexcept
				: 	generator( rand_dev() ),
					distr( 0, 100 ) {
		}

	public:

		/* Copy and move constructors are disabled */

		DBConnector( const DBConnector& orig ) = delete;
		DBConnector( DBConnector&& orig ) noexcept = delete;

		virtual ~DBConnector();

		/* Copy and move assignment operators are also disabled */

		DBConnector& operator=( const DBConnector& ) = delete;
		DBConnector& operator=( DBConnector&& ) = delete;

		/**
		 * Gets a connection to access reference counters
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to reference counters
		 */
		virtual RefcountDB* refcountDB() noexcept ( false ) = 0;

		/**
		 * Gets a connection to access locks
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to locks
		 */
		virtual LockDB* lockDB() noexcept ( false ) = 0;

		/**
		 * Gets a connection to check sessions
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to locks
		 */
		virtual SessionDB* sessionDB() noexcept ( false ) = 0;

		/**
		 * Gets a connection to access studies
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to the studies
		 */
		virtual StudyDB* studyDB() noexcept ( false ) = 0;

		/**
		 * Gets a connection to access objects
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to the objects
		 */
		virtual ObjectDB* objectDB() noexcept ( false ) = 0;

		/**
		 * Gets a connection to access several different tables
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a connection to access several different tables
		 */
		virtual MiscDB* miscDB() noexcept ( false ) = 0;

		/**
		 * Creates a new DB Connector.
		 * The connector should be released by the caller.
		 *
		 * @param fromThread true indicates to get a DB connector for a thread
		 *
		 * @return the DB connector.
		 */
		static DBConnector* createConnector( bool fromThread = false ) noexcept;

		/**
		 * Creates a new raw query to execute queries assembled dynamically.
		 * The caller shall take care to destroy the query when no longer needed.
		 *
		 * @throw DBException if an DB error occurs
		 *
		 * @return a new, empty raw query
		 */
		virtual RawQuery* createRawQuery() noexcept ( false ) = 0;

		/**
		 * Forks a new DBConnector object for a thread.
		 * Intended to be called from the thread,
		 * The caller shall take care to destroy the DBConnector when no longer needed.
		 *
		 * @return a new DBConnector initialized for a thread
		 */
		virtual DBConnector* forThread() noexcept = 0;

		/**
		 * Resets this connector object.
		 * The intention is to reset the connector - close the connections and drop any cached information -
		 * after an unrecoverable error, for example when a connection is lost.
		 */
		virtual void reset() noexcept = 0;


		inline void randomSleep() {
			std::this_thread::sleep_for( std::chrono::milliseconds( distr( generator ) ) );
		}

		inline size_t maxRetry() {
			return 20;
		}
};


/* ********************************************
 *
 * Implementation
 *
 * ********************************************/


class RefcountDBSql;
class StudyDBSql;
class ObjectDBSql;
class MiscDBSql;
class LockDBSql;
class SessionDBSql;

class PSSupplier;
class RawPSSupplier;
class CachedPSSupplier;
class PSCachedStatement;


/**
 * MySQL implementation of the DB Connector interface
 */
class DBConnectorSql : public DBConnector {
		// allow access to private constructors
		friend DBConnector* DBConnector::createConnector( bool );

	private:

		/**
		 * Creates a new DBConnectorSql object.
		 * When the formThread parameter is provided and true the MySQL library will be initialized for a thread.
		 * Otherwise the MySQL library will be initialized for a man process.
		 *
		 * @param fromThread [optional] indicates whether to create the connector for a thread or a main process
		 */
		DBConnectorSql( bool fromThread = false ) noexcept;

		/**
		 * Constructor to create a child connector for a thread
		 *
		 * @param parent the original database connector object
		 */
		DBConnectorSql( Sql* parent ) noexcept;

	public:

		/* Copy and move constructors are disabled */

		DBConnectorSql( const DBConnectorSql& orig ) = delete;
		DBConnectorSql( DBConnectorSql&& orig ) noexcept = delete;

		virtual ~DBConnectorSql();

		/* Copy and move assignment operators are also disabled */

		DBConnectorSql& operator=( const DBConnectorSql& ) = delete;
		DBConnectorSql& operator=( DBConnectorSql&& ) = delete;

		/**
		 * @copydoc DBConnector::refcountDB()
		 */
		virtual RefcountDB* refcountDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::lockDB()
		 */
		virtual LockDB* lockDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::sessionDB()
		 */
		virtual SessionDB* sessionDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::studyDB()
		 */
		virtual StudyDB* studyDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::objectDB()
		 */
		virtual ObjectDB* objectDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::miscDB()
		 */
		virtual MiscDB* miscDB() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::createRawQuery()
		 */
		virtual RawQuery* createRawQuery() noexcept ( false ) override;

		/**
		 * @copydoc DBConnector::forThread()
		 */
		virtual DBConnector* forThread() noexcept override;

		/**
		 * Gets a connection to MySQL.
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @attention This method is only intended for testing.
		 * Should never be used from production code.
		 */
		inline SqlConnection* getConnectionEx() noexcept ( false ) {
			if ( m_connection == NULL ) {
				m_connection = m_pDB->getConnectionEx();
			}
			return m_connection;
		}

		/**
		 * Prepares and executes a @code{select} query.
		 *
		 * @param query the query to execute
		 * @param bindings parameter and result bindings
		 *
		 * @throw DBException if an error occurs
		 *
		 * @return the sql statement to access the result set of the query
		 */
		PSCachedStatement executeQuery( const std::string& query, QueryBindings& bindings ) noexcept ( false );

		/**
		 * Prepares and executes an @code{update} query.
		 * An @code{update} query modifies (inserts/updates/deletes) records in the tables.
		 *
		 * @param query the query to execute
		 * @param bindings parameter bindings
		 * @param logError controls whether SQL errors to be logged
		 *
		 * @throw DBException if an error occurs
		 *
		 * @return the number of affected rows
		 */
		my_ulonglong executeUpdate( const std::string& query, QueryBindings& bindings, bool logError = true ) noexcept ( false );

		/**
		 * Executes the prepared statement of a @code{select} query.
		 *
		 * @param ps a supplier providing the prepared statement for the query
		 * @param bindings parameter and result bindings
		 *
		 * @throw DBException if an error occurs

		 * @return the sql statement to access the result set of the query
		 */
		PSCachedStatement executeQuery( PSSupplier& ps, QueryBindings& bindings ) noexcept ( false );

		/**
		 * Prepares and executes an @code{update} query.
		 * An @code{update} query modifies (inserts/updates/deletes) records in the tables.
		 *
		 * @param ps a supplier providing the prepared statement for the query
		 * @param bindings parameter and result bindings
		 * @param logError controls whether SQL errors to be logged
		 *
		 * @throw DBException if an error occurs
		 *
		 * @return the number of affected rows
		 */
		my_ulonglong executeUpdate( PSSupplier& ps, QueryBindings& bindings, bool logError = true ) noexcept ( false );

		/**
		 * @copydoc DBConnector::reset()
		 */
		virtual void reset() noexcept override;

		/**
		 * Gets the maximum number of threads cached for this database connector
		 * @return the maximum thread number for the cache
		 */
		inline size_t getMaxCachedThreads() const noexcept {
			return maxCachedThreads;
		}

		/**
		 * Sets the maximum number of threads cached for this database connector
		 * @param maxCachedThreads  the maximum thread number for the cache
		 */
		void setMaxCachedThreads( size_t maxCachedThreads ) noexcept {
			this->maxCachedThreads = maxCachedThreads;
		}

	private:

		/// The MySQL database engine
		Sql* m_pDB = NULL;
		/// MySQL database connection
		SqlConnection* m_connection = NULL;

		/// MySQL implementation of the reference counter database interface
		RefcountDBSql* m_pRefcountDB = NULL;
		/// MySQL implementation of the study work list database interface
		StudyDBSql* m_pStudyDB = NULL;
		/// MySQL implementation of the objects database interface
		ObjectDBSql* m_pObjectDB = NULL;
		/// MySQL implementation of the database interface to miscellaneous tables
		MiscDBSql* m_pMiscDB = NULL;
		/// MySQL implementation of the lock database interface
		LockDBSql* m_pLockDB = NULL;
		/// MySQL implementation of the session database interface
		SessionDBSql* m_pSessionDB = NULL;

		/// Mutex for locking new refcountDB creation
		std::mutex refcountMutex;
		/// Mutex for locking new studyDB creation
		std::mutex studyMutex;
		/// Mutex for locking new objectDB creation
		std::mutex objectMutex;
		/// Mutex for locking new miscDB creation
		std::mutex miscMutex;
		/// Mutex for locking new lockDB creation
		std::mutex lockMutex;
		/// Mutex for locking new sessionDB creation
		std::mutex sessionMutex;

		/// The maximum number of threads to be stored in the cache
		size_t maxCachedThreads;
};


/**
 * A bundle representing a MySQL prepared statement query.
 * Used by PSCache to prepare and cache the prepared statements
 */
class PsQuery {
		/// Identifies the query (index within the cache)
		size_t id;

		/// a function supplying the query to prepare
		std::function<std::string()> supplier;

		/// count of parameters in the query
		size_t params;

		/// count of columns in the result
		size_t columns;

	public:

		/**
		 * Default constructor to create an empty PSQuery object
		 */
		PsQuery();

		/* Copy and move constructors are disabled */

		PsQuery( const PsQuery& orig ) = delete;
		PsQuery( PsQuery&& orig ) = delete;

		~PsQuery();

		/* Copy and move assignment operators are also disabled */

		PsQuery& operator=( const PsQuery& ) = delete;
		PsQuery& operator=( PsQuery&& ) = delete;

		/**
		 * Fills in the PSQuery object
		 *
		 * @param id identifies the query
		 * @param supplier a function providing the query
		 * @param params count of parameters in the query
		 * @param columns count of result columns in the query
		 */
		inline void set( size_t id, std::function<std::string()>& supplier, size_t params, size_t columns ) {
			this->id = id;
			this->supplier = supplier;
			this->params = params;
			this->columns = columns;
		}

		/**
		 * Gets the query identifier.
		 * @return the query identifier
		 */
		inline const size_t getId() const {
			return id;
		}

		/**
		 * Gets the query to prepare
		 * @return the query
		 */
		inline std::string getQuery() const {
			return supplier();
		}

		/**
		 * Gets the number of parameters in the query.
		 * @return number of parameters
		 */
		inline const size_t getParamsNum() const {
			return params;
		}

		/**
		 * Gets the number of return columns in the query.
		 * @return number of return columns
		 */
		inline const size_t getColumnsNum() const {
			return columns;
		}

		/**
		 * Creates a new QueryBindings object for this query to bind parameters and results.
		 * It's the responsibility of the caller to release (delete) the QueryBindings object when no longer needed.
		 *
		 * @throw DBException if there is not enough memory to create the QueryBindings object.
		 * @return Pointer to the new QueryBindings object.
		 */
		inline QueryBindings* mkBindings() const noexcept ( false ) {
			try {
				return new QueryBindings( params, columns );
			}
			catch ( ... ) {
				throw DBException( "Not enough memory" );
			}
		}
};


/**
 * Cache of prepared statements created for a particular thread
 */
class StatementCache {

	public:
		/**
		 * Constructs a new @{StatementCache} object using a particular database connector
		 * and cache size
		 *
		 * @param db database connector
		 * @param size the number of statements to be stored in the cache
		 */
		StatementCache( DBConnectorSql& db, size_t size );

		/**
		 * Destroys @code{this} statement cache
		 */
		~StatementCache();

		/**
		 * Resets the thread's PS cache.
		 * Drops all prepared statements currently in the cache.
		 */
		void reset();

		/**
		 * Gets a prepared statement for the query for the current thread.
		 * @copydoc PSQuery::getPS(const PsQuery&)
		 */
		SqlStatement* getPS( const PsQuery& query ) noexcept ( false );

	private:

		/// MySQL connector
		DBConnectorSql& db;

		/// cache size (the number of statements to be stored in the cache)
		size_t size;

		/// the cache
		SqlStatement** cache;
};


/**
 * A helper class to prevent the cached statement go away when
 * another (newer) thread is dropping the current thread from the cache.
 *
 * Keeps a shared pointer to the local thread's statement cache to avoid it from being gone
 * while the @code{SqlStatement} is used.
 *
 * For single-use (or non-cached) statements the shared pointer remains empty.
 */
class PSCachedStatement {

	public:

		/**
		 * Constructs a new @code{PSCachedStatement} for a single-use statement that is actually not cached,
		 * @param sqlStatement the single-use statement
		 */
		PSCachedStatement( SqlStatement* sqlStatement )
				: sqlStatement( sqlStatement ) {
		}

		/**
		 * Constructs a new @code{PSCachedStatement} for a cached statement,
		 * @param sqlStatement the cached statement
		 * @param statementCache the cache storing the statement
		 */
		PSCachedStatement( SqlStatement* sqlStatement, std::shared_ptr<StatementCache> statementCache )
				: 	sqlStatement( sqlStatement ),
					statementCache( statementCache ) {
		}

		/* Use default copy and move constructors */

		PSCachedStatement( const PSCachedStatement& orig ) = default;
		PSCachedStatement( PSCachedStatement&& orig ) = default;

		~PSCachedStatement() {
		}

		/* Use default copy and move assignment operators */

		PSCachedStatement& operator=( const PSCachedStatement& ) = delete;
		PSCachedStatement& operator=( PSCachedStatement&& ) = delete;

		/**
		 * Gets the Sql Statement
		 * @return the Sql statement
		 */
		inline SqlStatement* getStatement() const {
			return sqlStatement;
		}

	private:
		/// the SQL statement
		SqlStatement* sqlStatement;
		/// the per thread PSCache for protecting while being used
		std::shared_ptr<StatementCache> statementCache;
};


/**
 * A cache of prepared statements of pre-registered queries
 *
 * <div>
 * 	Usage:
 * 		<ul>
 * 		<li>Register all queries using registerQuery()</li>
 * 		<li>call init()</li>
 * 		<li>use the prepared statements by getPS()</li>
 * 		</ul>
 * </div>
 */
class PSCache {

		friend class CachedPSSupplier;

	public:
		virtual ~PSCache();

		/**
		 * Resets the cache.
		 * Drops all prepared statements currently in the cache.
		 */
		void reset();

		/**
		 * Creates a new PSCache object.
		 *
		 * @attention
		 * The base PSQuery object should not be created directly.
		 * This constructor is for the derived classes.
		 *
		 * @param db MySQL connection
		 */
		PSCache( DBConnectorSql& db );

	protected:

		/**
		 * Gets a prepared statement for the query.
		 * Returns the prepared statement from the cache - or if it does not yet exists - create it and puts into the cache.
		 *
		 * @param query the query
		 *
		 * @throw SqlException if an SQL error occurs while preparing the query
		 *
		 * @return the prepared statement
		 */
		PSCachedStatement getPS( const PsQuery& query ) noexcept ( false );

		/**
		 * Registers a query to the cache.
		 * Queries should be added before init is called
		 *
		 * @param psQuery  an empty prepared statement query object to fill and register
		 * @param querySupplier    a function (or lambda) supplying the SQL query to be prepared and executed
		 * @param numParams number of parameters in the query (for binding)
		 * @param numColumns number of return columns in the query (for binding)
		 */
		inline void registerQuery( PsQuery& psQuery, std::function<std::string()>&& querySupplier, size_t numParams,
				size_t numColumns ) {
			psQuery.set( cacheSize++, querySupplier, numParams, numColumns );
		}

		/**
		 * Initializes the cache with the registered queries.
		 */
		void init();

		/**
		 * Prepares and executes a @code{select} query.
		 * A @code{select} query does not modify any records, but returns data from the tables.
		 *
		 * @param query the query to execute
		 * @param bindings parameter and result bindings
		 * @throw DBException if an error occurs
		 * @return the sql statement to access the result set of the query
		 */
		PSCachedStatement executeQuery( const PsQuery& query, QueryBindings& bindings ) noexcept ( false );

		/**
		 * Prepares and executes an @code{update} query.
		 * An @code{update} query modifies (inserts/updates/deletes) records in the tables.
		 *
		 * @param query the query to execute
		 * @param bindings parameter bindings
		 * @throw DBException if an error occurs
		 * @return the number of affected rows
		 */
		my_ulonglong executeUpdate( const PsQuery& query, QueryBindings& bindings, bool logError = true ) noexcept ( false );

	protected:
		/// MySQL connector
		DBConnectorSql& db;
		/// the per thread statement caches
		std::unique_ptr<erad::imutils::ThreadCache<StatementCache>> cache;

	private:
		/// The count of prepared statements stored for a single thread
		size_t cacheSize;
};


/**
 * Abstract base class (interface) to create and delete prepared statements to execute SQL queries.
 *
 * Prepared statement suppliers are used internally by the SQL connector classes.
 */
class PSSupplier {
		friend class DBConnectorSql;

	protected:
		/* Use the default constructor */

		PSSupplier() = default;

		/* Copy and move constructors are disabled */

		PSSupplier( const PSSupplier& orig ) = delete;
		PSSupplier( PSSupplier&& orig ) = delete;

		virtual ~PSSupplier();

		/* Copy and move assignment operators are also disabled */

		PSSupplier& operator=( const PSSupplier& ) = delete;
		PSSupplier& operator=( PSSupplier&& ) = delete;

		/**
		 * Creates the prepared statement.
		 *
		 * @throw SqlException if an error occurs
		 *
		 * @return the prepared statement to execute
		 */
		virtual PSCachedStatement create() noexcept ( false ) = 0;

		/**
		 * Releases a prepared statement returned by a @code{create()}.
		 *
		 * @param ps the prepared statement to release
		 */
		virtual void release( SqlStatement* ps ) noexcept = 0;
};


/**
 * PS supplier to generate a prepared statement for a single use.
 *
 * Used internally by DBConnectorSql
 */
class RawPSSupplier : public PSSupplier {

		friend class DBConnectorSql;

	protected:

		/**
		 * Creates a RawPSSupplier object
		 *
		 * @param db SQL database connector
		 * @param query the query to prepare
		 */
		RawPSSupplier( DBConnectorSql& db, const std::string& query );

		/* Default, copy and move constructors are disabled */

		RawPSSupplier() = delete;
		RawPSSupplier( const RawPSSupplier& orig ) = delete;
		RawPSSupplier( RawPSSupplier&& orig ) = delete;

		virtual ~RawPSSupplier();

		/* Copy and move assignment operators are also disabled */

		RawPSSupplier& operator=( const RawPSSupplier& ) = delete;
		RawPSSupplier& operator=( RawPSSupplier&& ) = delete;

		/**
		 * @copydoc PSSupplier::create()
		 *
		 * Always creates a new prepared statement
		 */
		virtual PSCachedStatement create() noexcept ( false ) override;

		/**
		 * @copydoc PSSupplier::release(SqlStatement*)
		 *
		 * Deletes the prepared statement
		 */
		virtual void release( SqlStatement* ps ) noexcept override;

	private:

		/// SQL database connector
		DBConnectorSql& db;
		/// the query to prepare
		const std::string& query;
};


/**
 * A PS supplier to create prepared statements and cache it for later reuse
 *
 * Used internally by PSCache
 */
class CachedPSSupplier : public PSSupplier {

		friend class PSCache;

	protected:

		/**
		 * Creates a new CachedPSSupplier object
		 *
		 * @param cache the prepared statement cache
		 * @param query the query to prepare
		 */
		CachedPSSupplier( PSCache& cache, const PsQuery& query );

		/* Default, copy and move constructors are disabled */

		CachedPSSupplier() = delete;
		CachedPSSupplier( const CachedPSSupplier& orig ) = delete;
		CachedPSSupplier( CachedPSSupplier&& orig ) = delete;

		virtual ~CachedPSSupplier();

		/* Copy and move assignment operators are also disabled */

		CachedPSSupplier& operator=( const CachedPSSupplier& ) = delete;
		CachedPSSupplier& operator=( CachedPSSupplier&& ) = delete;

		/**
		 * @copydoc PSSupplier::create()
		 *
		 * Returns the prepared statement from the cache if already exist,
		 * or generate a new prepared statement and puts to the cache
		 */
		virtual PSCachedStatement create() noexcept ( false ) override;

		/**
		 * @copydoc PSSupplier::release(SqlStatement*)
		 * Does nothing
		 */
		virtual void release( SqlStatement* ps ) noexcept override;

	private:

		/// The prepared statement cache
		PSCache& cache;
		/// The query to prepare
		const PsQuery& query;
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_DBCONNECTOR_H_ */
