/*
 * Component:  utils
 *
 * Purpose:
 *    Logging implementation using syslog
 *
 */
#ifndef IM_LOGGER_H_INCLUDED
#define IM_LOGGER_H_INCLUDED

#include <stdarg.h>
#include <string>


typedef enum {
	LOGMODE_OPER_INFO = 0,
	LOGMODE_OPER_ERROR = 1,
	LOGMODE_INFO = 2,
	LOGMODE_ERROR = 3,
	LOGMODE_DEBUG = 4
} LOG_MODE;


typedef enum {
	LOGFAC_OPER,
	LOGFAC_NORMAL,
	LOGFAC_NULL
} LOG_FACILITY;


class IMLogger {
	private:

		/** Facility to be used for logging {LOGFAC_OPER|LOGFAC_NORMAL} */
//		static LOG_FACILITY m_lastFacility;
		LOG_FACILITY m_lastFacility;

		/** Loggin tag */
		char* m_tag;

		/**
		 * Open syslog logger facility
		 *
		 * @param logFacility    logger facility to be opened
		 */
		void openLog( LOG_FACILITY logFacility );

	protected:
		/**
		 * Create log entry. Possible values for mode:
		 *
		 *	LOGMODE_OPER_INFO  : create entry in oper_info.log
		 *  LOGMODE_OPER_ERROR : create entry in oper_error.log
		 *  LOGMODE_INFO       : create entry in info.log
		 *  LOGMODE_ERROR      : create entry in error.log
		 *  LOGMODE_DEBUG      : create entry in debug.log
		 *
		 * @param mode    Mode of the log entry to be created
		 * @param mask    Mask for log message
		 * @param ap      Arguments to be replaced in mask
		 */
		void log( LOG_MODE mode, const char* mask, va_list& ap );

	public:

		/**
		 * Create new Logger object with logging tag
		 *
		 * @param tag  Logger tag to be used in log messages
		 */
		IMLogger( const char* tag );

		/**
		 * Destructor for destroying Logger objects
		 */
		~IMLogger();

		/**
		 * Create log entry in debug.log
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void debugLog( const char* mask, ... );

		/**
		 * Create log entry in info.log
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void infoLog( const char* mask, ... );

		/**
		 * Create log entry in oper_info.log
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void operInfoLog( const char* mask, ... );

		/**
		 * Create log entry in error.log
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void errorLog( const char* mask, ... );

};

#endif // IM_LOGGER_H_INCLUDED
