/*
 * Raw Query
 *
 * Interface and implementation to execute raw queries
 *
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_RAW_QUERY_H_
#define IMUTILS_INCLUDE_IMUTILS_RAW_QUERY_H_

#include "imutils/DBConnector.h"
#include "imutils/Sql.h"
#include <string>
#include <list>
#include <functional>
#include <regex>

namespace erad::db {


class JoinTable {
	public:

		// Join type (left, straight, natural, ...)
		std::string type;
		/// the table to join to the primary table
		std::string table;
		/// the field in the table to join on
		std::string joinField;
		/// the name of the primary table (especially when there are multiple tables)
		std::string primaryTable;
		/// the join field in the primary table
		std::string primaryField;

		JoinTable( const std::string& type, const std::string& table, const std::string& joinField, const std::string& primaryTable,
				const std::string& primaryField ) noexcept;

		JoinTable( const JoinTable& orig ) = delete;
		JoinTable( JoinTable&& orig ) = delete;
		virtual ~JoinTable();

		JoinTable& operator=( const JoinTable& ) = delete;
		JoinTable& operator=( JoinTable&& ) = delete;
};


enum class Comp {
	/// Equal to
	EQ,
	/// Not equal to
	NE,
	/// Greater than
	GT,
	/// Greater than or equal to
	GE,
	/// Less than
	LT,
	/// Less than or equal to
	LE
};


/**
 * Helper class to parse DICOM TM values and convert to SQL format.
 */
class TimeValue {
	public:
		/**
		 * Creates a new <code>TimeValue</code> object by parsing a DICOM TM value
		 *
		 * @param str the DICOM TM value
		 */
		TimeValue( const std::string& str );
		~TimeValue();

		/**
		 * Checks whether the time string is valid, at least syntactically matches to the DICOM TM VR.
		 * @return true if the time string is valid
		 */
		inline bool isValid() const {
			return valid;
		}

		/**
		 * Appends this TimeVale to the provided string in SQL format
		 * @param str the output string
		 * @return the output string
		 */
		inline std::string& appendSqlTime( std::string& str ) const {
			str.append( hours ).append( ":" )
					.append( minutes.empty() ? "00" : minutes ).append( ":" )
					.append( seconds.empty() ? "00" : seconds );

			// Ignore fragments as there are no fragments stored in the database
			// and in MySQL HH:MM:SS does not match to the HH:MM:SS.000000 <= time <= HH:MM:SS.999999 range
//			if ( !fragments.empty() ) {
//				str.append( "." ).append( fragments );
//			}
			return str;
		}

	private:

		/** The hours extracted from the TM value */
		std::string hours;
		/** The minutes extracted from the TM value */
		std::string minutes;
		/** The seconds extracted from the TM value */
		std::string seconds;
		/** The fragments (microseconds) extracted from the TM value */
		std::string fragments;

		/** flag indicating whether the TM value is valid */
		bool valid;

		/** Regular expression to parse the TM value */
		static std::regex parseRegex;
};


class RawQuery {

	public:
		virtual ~RawQuery();

		/**
		 * Adds a field to the result field list of a @code{select} query.
		 *
		 * @param field the field to be added to the query
		 *
		 * @return the index of the field
		 */
		virtual size_t addField( const std::string& field ) noexcept = 0;

		/**
		 * Adds a field to the result field list of a @code{join} query.
		 * The field name will be prefixed with the table to avoid ambiguity.
		 *
		 * @param field the field to be added to the query
		 * @param table table prefix for the field.
		 *
		 * @return the index of the field
		 */
		virtual size_t addField( const std::string& field, const std::string& table ) noexcept = 0;

		/**
		 * Adds a field to the GROUP BY clause of a @code{select} query.
		 *
		 * @param field the field to be added
		 */
		virtual void addGroupByField( const std::string& field ) noexcept = 0;

		/**
		 * Adds a field to the GROUP BY clause of a @code{join} query.
		 * The field name will be prefixed with the table to avoid ambiguity.
		 *
		 * @param field the field to be added
		 * @param table table prefix for the field.
		 */
		virtual void addGroupByField( const std::string& field, const std::string& table ) noexcept = 0;

		/**
		 * Adds a field to the ORDER BY clause of a @code{select} query.
		 *
		 * @param field the field to be added
		 */
		virtual void addOrderByField( const std::string& field ) noexcept = 0;

		/**
		 * Adds a field to the ORDER BY clause of a @code{join} query.
		 * The field name will be prefixed with the table to avoid ambiguity.
		 *
		 * @param field the field to be added
		 * @param table table prefix for the field.
		 */
		virtual void addOrderByField( const std::string& field, const std::string& table ) noexcept = 0;

		/**
		 * Sets the table for a @code{select} query, or the primary table for a @code{join} query
		 *
		 * @param table the SQL table
		 */
		virtual void setTable( const std::string& table ) noexcept =0;

		/**
		 * Sets a secondary table for a @code{join} query.
		 *
		 * @param type the join type (natural, left, ...)
		 * @param table the SQL table to join
		 * @param primaryField field in the primary table to use in the join condition
		 * @param joinField field in the secondary table to use in the join condition
		 */
		virtual void addJoinTable( const std::string& type, const std::string& table, const std::string& primaryField,
				const std::string& joinField ) noexcept =0;

		/**
		 * Adds a filter to the WHERE clause of a @code{select} query for a numeric field.
		 *
		 * @param dbName the field to filter on
		 * @param oper filter operation
		 * @param value filter value
		 */
		virtual void addFilter( const std::string& dbName, const Comp& oper, int value ) noexcept =0;

		/**
		 * Adds a filter to the WHERE clause of a @code{select} query for a string field.
		 *
		 * @param dbName the field to filter on
		 * @param oper filter operation
		 * @param value filter value
		 */
		virtual void addFilter( const std::string& dbName, const Comp& oper, const std::string& value ) noexcept =0;

		/**
		 * Adds a @code{like} filter to the WHERE clause of a @code{select} query for a multi-value field.
		 * Multi-value fields may have multiple values separated by backslash.
		 *
		 * @param dbName the SQL field
		 * @param value a single value to search
		 * @param caseSensitive true indicates to search the value case sensitively
		 * @param emptyMatch true indicates to also consider empty database values as a match
		 */
		virtual void addMultiValueLikeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept = 0;

		/**
		 * Adds an @code{is} filter to the WHERE clause of a @code{select} query for a field.
		 *
		 * @param dbName the SQL field
		 * @param value the value to filter on
		 * @param caseSensitive true indicates to search the value case sensitively
		 * @param emptyMatch true indicates to also consider empty database values as a match
		 */
		virtual void addSingleClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept = 0;

		/**
		 * Adds an @code{range} filter to the WHERE clause of a @code{select} query for field.
		 *
		 * @param dbName the SQL field
		 * @param value the range to filter on (expected to be in [@code{from}]-[@code{to}] format)
		 * @param caseSensitive true indicates to search the value case sensitively
		 * @param emptyMatch true indicates to also consider empty database values as a match
		 */
		virtual void addRangeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept = 0;

		/**
		 * Adds an @code{in} filter to the WHERE clause of a @code{select} query for field.
		 *
		 * @param dbName the SQL field
		 * @param value backslash separated list of individual values to filter on
		 */
		virtual void addInClause( const std::string& dbName, const std::string& value ) noexcept = 0;

		/**
		 * Adds a @code{like} filter to the WHERE clause of a @code{select} query for a single-value field.
		 *
		 * @param dbName the SQL field
		 * @param value the @code{like} pattern to be used in the filter
		 * @param caseSensitive true indicates to search the value case sensitively
		 * @param emptyMatch true indicates to also consider empty database values as a match
		 */
		virtual void addLikeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept =0;

		/**
		 * Constructs an SQL clause for a date/time field and adds the clause
		 * to the WHERE clause of a @code{select} query.
		 *
		 * @param dbName the DATETIME field
		 * @param date a date range filter string
		 * @param time a time range filter string
		 * @param tzOffset time zone offset for the time range filter
		 * @param combine true indicates to combine the date and time range filters into a single date/time range
		 * @param emptyMatch true indicates to also consider empty database values as a match
		 */
		virtual void addDateTimeClause( const std::string& dbName, const std::string& date, const std::string time,
				const char* tzOffset, bool combine, bool emptyMatch ) noexcept =0;

		/**
		 * Gets the number of return fields in a @code{select} or @code{join} query.
		 *
		 * @return the number of return fields
		 */
		virtual size_t countFields() noexcept =0;

		/**
		 * Checks if the query has a where clause.
		 *
		 * @return true if and only if the query has a WHERE clause
		 */
		virtual bool hasWhere() noexcept =0;

		/**
		 * Creates the SELECT query from the query parts set for this query object.
		 */
		virtual void compile() noexcept = 0;

		/**
		 * Gets the SELECT query string created by compile().
		 *
		 * @return the SELECT query string (if already compiled)
		 */
		virtual std::string str() noexcept = 0;

		/**
		 * Sets whether to return all or only distinct values
		 *
		 * @param distinct set to true to return only distinct values
		 */
		virtual void setDistinct( bool distinct ) noexcept =0;

		/**
		 * Executes the query.
		 * Handles 'connection gone' conditions by reopening the connection.
		 *
		 * If a timeout is provided, then fails if the query does not complete within the timeout.
		 *
		 * @param timeout timeout in seconds (default: 0 - no timeout)
		 *
		 * @throw DBException if an error occurs
		 */
		virtual void executeEx( size_t timeout = 0 ) noexcept ( false ) = 0;

		/**
		 * Gets the error message - if executing the query failed
		 *
		 * @return the error message
		 */
		virtual const std::string& getError() noexcept =0;

		/**
		 * Checks whether executing the query was interrupted (timed out) or not.
		 *
		 * @return true if and only if executing the query was interrupted
		 */
		virtual bool isInterrupted() noexcept =0;

		/**
		 * Takes the next record in the query result set.
		 *
		 * @return true if and only if there was a next record.
		 */
		virtual bool nextRecord() noexcept =0;

		/**
		 * Checks if the field on the provided index is NULL
		 *
		 * @param fld the index of the field
		 *
		 * @throw std::out_of_range if no such field
		 *
		 * @return true if and only if the field is NULL
		 */
		virtual bool isNull( size_t fld ) noexcept ( false ) = 0;

		/**
		 * Gets the value field on the provided index
		 *
		 * @param fld the index of the field
		 *
		 * @throw std::out_of_range if no such field
		 *
		 * @return the field value
		 */
		virtual const std::string& getValue( size_t fld ) noexcept ( false ) = 0;
};


class RawQueryDB : public RawQuery {
		/// DB connector to get a DB connection
		DBConnectorSql& db;

		/// DB connection to execute the query
		SqlConnection* connection;

		/// the select fields
		std::vector<std::string> fields;

		/// the primary table of the query
		std::string primaryTable;

		/// join tables - if any
		std::list<JoinTable> joinTables;

		/// the where clause
		std::string where;

		/// controls whether to select all or only distinct values
		bool distinct;

		/// the entire query
		std::string query;

		/// 'group by' fields - if any
		std::list<std::string> groupBy;

		/// 'order by' fields - if any
		std::list<std::string> orderBy;

		SqlResult* result;

		std::vector<SqlField> record;

	public:
		/**
		 * Creates a new RawQueryDB object
		 *
		 * @param db DB connector
		 */
		RawQueryDB( DBConnectorSql& db ) noexcept;

		/* Default, copy and move constructors are disabled */

		RawQueryDB() = delete;
		RawQueryDB( const RawQueryDB& orig ) = delete;
		RawQueryDB( RawQueryDB&& orig ) = delete;

		virtual ~RawQueryDB();

		/* Copy and move assignment operators are also disabled */

		RawQuery& operator=( const RawQuery& ) = delete;
		RawQuery& operator=( RawQuery&& ) = delete;

		/**
		 * @copydoc RawQuery::addField(const std::string&)
		 */
		virtual size_t addField( const std::string& field ) noexcept override;

		/**
		 * @copydoc RawQuery::addField(const std::string&,const std::string&)
		 */
		virtual size_t addField( const std::string& field, const std::string& table ) noexcept override;

		/**
		 * @copydoc RawQuery::addGroupByField(const std::string&)
		 */
		virtual void addGroupByField( const std::string& field ) noexcept override;

		/**
		 * @copydoc RawQuery::addGroupByField(const std::string&,const std::string&)
		 */
		virtual void addGroupByField( const std::string& field, const std::string& table ) noexcept override;

		/**
		 * @copydoc RawQuery::addOrderByField(const std::string&)
		 */
		virtual void addOrderByField( const std::string& field ) noexcept override;

		/**
		 * @copydoc RawQuery::addOrderByField(const std::string&,const std::string&)
		 */
		virtual void addOrderByField( const std::string& field, const std::string& table ) noexcept override;

		/**
		 * @copydoc RawQuery::setTable(const std::string&)
		 */
		virtual void setTable( const std::string& table ) noexcept override;

		/**
		 * @copydoc RawQuery::addJoinTable(const std::string&,const std::string&,const std::string&,const std::string&)
		 */
		virtual void addJoinTable( const std::string& type, const std::string& table, const std::string& primaryField,
				const std::string& joinField ) noexcept override;

		/**
		 * @copydoc RawQuery::addFilter(const std::string&,const Comp&,int)
		 */
		virtual void addFilter( const std::string& dbName, const Comp& oper, int value ) noexcept override;

		/**
		 * @copydoc RawQuery::addFilter(const std::string&,const Comp&,const std::string&)
		 */
		virtual void addFilter( const std::string& dbName, const Comp& oper, const std::string& value ) noexcept override;

		/**
		 * @copydoc RawQuery::addMultiValueLikeClause(const std::string&,const std::string&,bool,bool)
		 */
		virtual void addMultiValueLikeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept override;

		/**
		 * @copydoc RawQuery::addSingleClause(const std::string&,const std::string&,bool,bool)
		 */
		virtual void addSingleClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept override;

		/**
		 * @copydoc RawQuery::addRangeClause(const std::string&,const std::string&,bool,bool)
		 */
		virtual void addRangeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept override;

		/**
		 * @copydoc RawQuery::addInClause(const std::string&,const std::string&)
		 */
		virtual void addInClause( const std::string& dbName, const std::string& value ) noexcept override;

		/**
		 * @copydoc RawQuery::addLikeClause(const std::string&,const std::string&,bool,bool)
		 */
		virtual void addLikeClause( const std::string& dbName, const std::string& value, bool caseSensitive,
				bool emptyMatch ) noexcept override;

		/**
		 * @copydoc RawQuery::addDateTimeClause(const std::string&,const std::string&,const std::string&,const char*,bool,bool)
		 */
		virtual void addDateTimeClause( const std::string& dbName, const std::string& date, const std::string time,
				const char* tzOffset, bool combine, bool emptyMatch ) noexcept override;

		/**
		 * @copydoc RawQuery::countFields()
		 */
		virtual size_t countFields() noexcept override;

		/**
		 * @copydoc RawQuery::hasWhere()
		 */
		virtual bool hasWhere() noexcept override;

		/**
		 * @copydoc RawQuery::compile()
		 */
		virtual void compile() noexcept override;

		/**
		 * @copydoc RawQuery::str()
		 */
		virtual std::string str() noexcept override;

		/**
		 * @copydoc RawQuery::setDistinct(bool)
		 */
		virtual void setDistinct( bool distinct ) noexcept override;

		/**
		 * @copydoc RawQuery::executeEx(size_t)
		 *
		 * Handles 'connection gone' conditions by reopening the connection.
		 */
		virtual void executeEx( size_t timeout = 0 ) noexcept ( false ) override;

		/**
		 * @copydoc RawQuery::getError()
		 */
		virtual const std::string& getError() noexcept override;

		/**
		 * @copydoc RawQuery::isInterrupted()
		 */
		virtual bool isInterrupted() noexcept override;

		/**
		 * @copydoc RawQuery::nextRecord()
		 */
		virtual bool nextRecord() noexcept override;

		/**
		 * @copydoc RawQuery::isNull(size_t)
		 */
		virtual bool isNull( size_t fld ) noexcept ( false ) override;

		/**
		 * @copydoc RawQuery::getValue(size_t)
		 */
		virtual const std::string& getValue( size_t fld ) noexcept ( false ) override;

	private:

		/**
		 * SQL encode a string for a query
		 *
		 * @param source the string to SQL encode
		 * @param dest output parameter for the encoded string
		 * @param isLike true indicates to encode the string for a @code{LIKE} clause
		 */
		void sqlEncode( const std::string& source, std::string& dest, bool isLike = false );

		/**
		 * Converts a date value to SQL format
		 * @param dateValue the date in 'YYYYmmdd' format
		 * @return the date value in SQL format
		 */
		std::string dateValue2Sql( const std::string& dateValue );

		/**
		 * Converts starting and ending dates in a DICOM date range to SQL date strings.
		 *
		 * @param dateRange the DICOM date range to convert
		 * @param posDash   position of dash in the DICOM date range
		 * @param from      output parameter for the start date in SQL format
		 * @param to        output parameter for the end date in SQL format
		 */
		bool convertDateRange2Sql( const std::string& dateRange, size_t posDash, std::string& from /* out */,
				std::string& to /* out */);

		/**
		 * Converts starting and ending times in a DICOM time range to SQL time strings.
		 *
		 * @param timeRange the DICOM time range to convert
		 * @param posDash   position of dash in the DICOM time range
		 * @param from      output parameter for the start time in SQL format
		 * @param to        output parameter for the end time in SQL format
		 */
		bool convertTimeRange2Sql( const std::string& timeRange, size_t posDash, std::string& from /* out */,
				std::string& to /* out */);

		/**
		 * Converts starting and ending date/times in a DICOM date/time ranges to SQL date/time strings.
		 *
		 * Used when Date-time matching is allowed in extended negotiation
		 *
		 * @param dateRange     the DICOM date range to convert
		 * @param posDateDash   position of dash in the DICOM date range
		 * @param timeRange     the DICOM time range to convert
		 * @param posTimeDash   position of dash in the DICOM time range
		 * @param from          output parameter for the start date/time in SQL format
		 * @param to            output parameter for the end date/time in SQL format
		 */
		bool convertDateTimeRange2Sql( const std::string& dateRange, size_t posDateDash, const std::string& timeRange,
				size_t posTimeDash, std::string& from /* out */, std::string& to /* out */);

		/**
		 * Appends a range filter expression for a date or time field to the WHERE clause.
		 *
		 * @param dbname the SQL field
		 * @param from the low value for the range
		 * @param to the high value for the range
		 * @param needQuotes true indicates to surround the values with a quote
		 * @param addAnd true indicates to prepend an @code{AND} to the expression
		 */
		void appendDateTimeRangeClause( const std::string& dbname, const std::string& from, const std::string& to, bool needQuotes,
				bool addAnd );

		/**
		 * Appends a simple filter expression for a date or time field to the WHERE clause.
		 *
		 * @param dbname the SQL field
		 * @param value the date or time value
		 * @param needQuotes true indicates to surround the values with a quote
		 */
		void appendDateTimeClause( const std::string& dbName, const std::string& value, bool needQuotes );

};

} // namespace

#endif // IMUTILS_INCLUDE_IMUTILS_RAW_QUERY_H_
