/*
 * TemplateFile.h
 *
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_TESTUTILS_H_
#define IMUTILS_INCLUDE_IMUTILS_TESTUTILS_H_

#include <gtest/gtest.h>

#include <string>
#include <map>
#include <list>
#include <regex>

namespace erad { namespace testing {

class TemplateMap : public std::map<std::string,std::string> {
public:
	TemplateMap();
	virtual ~TemplateMap();

	bool applyTo( const std::string & templateFile, const std::string & outputFile ) const;
	bool applyTo( const std::string & templateDirectory, const std::string & outputDirectory, const std::regex & pattern ) const;
};


class PackageFile {
public:
	PackageFile();
	PackageFile( const std::string & path );
	virtual ~PackageFile();

	bool unpack();

	inline void setPath( const std::string & path ) {
		this->path = path;
	}

	inline const std::string & getPath() const {
		return path;
	}

	inline bool isUnpacked() const {
		return unpacked;
	}

	friend std::ostream& operator<<( std::ostream& os, const PackageFile & obj ) {
		return os << obj.path << " (" << std::boolalpha << obj.unpacked << ")";
	}


protected:
	std::string path;
	bool unpacked;
};


class Environment : public ::testing::Environment {
public:
	Environment();
	virtual ~Environment();

	inline void setSourceDir( const std::string & sourceDir ) {
		this->sourceDir = sourceDir;
	}

	inline void setWorkDir( const std::string & workDir ) {
		this->workDir = workDir;
	}

	inline const std::string & getSourceDir() const {
		return sourceDir;
	}

	inline const std::string & getWorkDir() const {
		return workDir;
	}

	virtual ::testing::AssertionResult checkEnv();

	friend std::ostream& operator<<( std::ostream& os, const Environment & env );

protected:
	std::string sourceDir;
	std::string workDir;
};

/**
 * List files from a directory matching to a regex pattern.
 *
 * When replace is provided, it will be applied to the regex pattern for the matching files
 * to add a formatted name to the list.
 *
 * It can be used - for example - to remove extension from the file name:
 *
 *     dir/
 *         apple.txt
 *         banana.jpg
 *         cherry.txt
 *         date.png
 *
 *     listFiles( dir, std::regex( "^(.*)\\.txt$" ), "$1" );
 *
 * would return all .txt files but without the extension:
 *
 *         apple
 *         cherry
 *
 */
std::list<std::string> listFiles( const std::string & directory, const std::regex & pattern, const char * replace = nullptr );

/**
 * List files from a directory recursivly matching to a regex pattern.
 *
 * When replace is provided, it will be applied to the regex pattern for the matching files
 * to add a formatted name to the list.
 *
 * It can be used - for example - to remove extension from the file name:
 *
 *     dir/dir/
 *         apple.txt
 *         banana.jpg
 *         cherry.txt
 *         date.png
 *
 *     listFiles( dir, std::regex( "^(.*)\\.txt$" ), "$1" );
 *
 * would return all .txt files but without the extension:
 *
 *         apple
 *         cherry
 *
 */
std::list<std::string> listFilesRecursivly( const std::string & directory, const std::regex & pattern, const char * replace = nullptr ,std::string relativeDir="");



/**
 * Deletes directory recursively
 */
void deleteDirRecursively( const std::string & dirName );


/**
 * Checks if a string is empty and returns an assertion result
 * Succeeds if and only if the string is empty.
 */
::testing::AssertionResult isEmpty( const std::string & str );


/**
 * Compares two files using system cmp command and returns an assertion result.
 * Succeeds if and only if the files are exactly the same.
 */
::testing::AssertionResult compareFiles( const std::string & outputFile, const std::string & expectedFile );


/**
 * Compares a string to the content of a file and returns an assertion result.
 * Succeeds if and only if the string and the content of the file are exactly the same.
 */
::testing::AssertionResult compareStrToFile( const std::string & str, const std::string & file );


/**
 * Checks if a file exists (using stat).
 * Succeeds if and only if the file exists
 */
bool fileExists( const std::string & path );


} } // namespace


#endif /* IMUTILS_INCLUDE_IMUTILS_TESTUTILS_H_ */
