/**
 * Component: pbdcm
 *
 * Description:
 *     Header file for PB Custom Fields class manipulating custom fields
 */

#ifndef PBCUSTOMPROPERTIES_H_
#define PBCUSTOMPROPERTIES_H_

#include <dcmtk/dcmdata/dctk.h>


/**
 * Namespace for DICOM utility functions and classes.
 * Legacy functions and classes intended to be moved here eventually
 */
namespace erad::pbdcm {


/**
 * Helper class to manipulate custom fields in an
 * existing PB Custom Fields sequence within a DICOM object.
 */
class CustomFieldsSeq {
	private:
		/** PB Custom Fields sequence */
		DcmSequenceOfItems* m_customFieldsSeq;

	public:
		/**
		 * Constructs a new `CustomFieldsSeq` object for an existing PB Custom Fields sequence element.
		 * @param seq the PB Custom Fields sequence element
		 */
		CustomFieldsSeq( DcmSequenceOfItems* seq );

		/**
		 * Gets the value for the custom defined field.
		 *
		 * @param fieldName  name of the custom defined field
		 *
		 * @return  a pair of a `string` and a `bool` value;
		 *          the `bool` value indicating whether the field was found, and
		 *          the `string` value is the value of the field when found
		 */
		std::pair<std::string, bool> getFieldValue( const std::string& fieldName ) const;


		/**
		 * Sets value for custom defined field.
		 * If value is NULL, the field will be deleted from the DICOM object
		 *
		 * @param fieldName   name of the custom defined field
		 * @param value       new value for the custom defined field
		 *
		 * @return true if and only if the new value was successfully set for the field
		 */
		bool setFieldValue( const std::string& fieldName, const std::string& value );

		/**
		 * Deletes a custom field
		 * @param fieldName the name of the custom field to delete
		 * @return `true` if the field did not exist or could be successfully deleted, or `false` if an error occurred.
		 */
		bool deleteFieldValue( const std::string& fieldName );

		/**
		 * Return the field names currently set in the DICOM object.
		 *
		 * @return the list of the custom defined field names
		 */
		std::vector<std::string> getFieldNames() const;
};


/**
 * Manipulating PB custom defined fields in DICOM objects
 *
 * Creates or deletes the PB Custom Fields sequence element in the DICOM objects
 * and insert/deletes custom fields to it.
 */
class CustomFields {
	private:

		/** Dataset of the DICOM file */
		DcmDataset & m_dataset;

		/** PB Custom Fields sequence */
		DcmSequenceOfItems * m_customFieldsSeq;

	public:

		/**
		 * Constructor for creating a new PB Custom Fields object
		 *
		 * @param dataset  dataset of the DICOM object
		 */
		CustomFields( DcmDataset & dataset );

		/**
		 * Returns value for the custom defined field
		 *
		 * @param fieldName  name of the custom defined field
		 *
		 * @return  a pair of a `string` and a `bool` value;
		 *          the `bool` value indicating whether the field was found, and
		 *          the `string` value is the value of the field when found
		 */
		std::pair<std::string, bool> getFieldValue( const std::string& fieldName ) const;

		/**
		 * Sets value for custom defined field.
		 * Inserts a PB Custom Fields sequence element to the object data set if it does not yet exist.
		 *
		 * @param fieldName   name of the custom defined field
		 * @param value       new value for the custom defined field
		 *
		 * @return true if and only if the new value was successfully set for the field
		 */
		bool setFieldValue( const std::string& fieldName, const std::string& value );


		/**
		 * Deletes a custom field.
		 * Also deletes the PB Custom Fields sequence element from the object data set
		 * if no more custom fields remained in the sequence.
		 *
		 * @param fieldName the name of the custom field to delete
		 * @return `true` if the field did not exist or could be successfully deleted, or `false` if an error occurred.
		 */
		bool deleteFieldValue( const std::string& fieldName );

		/**
		 * Return the field names currently set in the DICOM object.
		 *
		 * @return the list of the custom defined field names
		 */
		std::vector<std::string> getFieldNames() const;
};


} // namespace



/**
 * Manipulating PB custom defined fields in DICOM objects
 */
class PBCustomFields {
	private:

		/** Dataset of the DICOM file */
		DcmDataset & m_dataset;

		/** PB Custom Fields sequence */
		DcmSequenceOfItems * m_customFieldsSeq;

	public:

		/**
		 * Constructor for creating a new PB Custom Fields object
		 *
		 * @param dataset  dataset of the DICOM object
		 */
		PBCustomFields( DcmDataset & dataset );

		/**
		 * Destructor for destroying the PB Custom Fields object
		 */
		virtual ~PBCustomFields();

		/**
		 * Returns value for the custom defined field
		 *
		 * @param fieldName  name of the custom defined field
		 *
		 * @return  field value or NULL if no such field in the DICOM object
		 */
		const char * getFieldValue( const char * fieldName ) const;

		/**
		 * Sets value for custom defined field.
		 * If value is NULL, the field will be deleted from the DICOM object
		 *
		 * @param fieldName   name of the custom defined field
		 * @param value       new value for the custom defined field
		 *
		 * @return true if and only if the new value was successfully set for the field
		 */
		bool setFieldValue( const char * fieldName, const char * value );

		/**
		 * Return the field names currently set in the DICOM object.
		 * Field names will be returned in the nameList parameter (will be cleared first).
		 *
		 * @param nameList  output parameter for the custom defined field names
		 */
		void fieldNames( OFList< OFString > & nameList ) const;
};

#endif /* PBCUSTOMPROPERTIES_H_ */
