/*
 * pbdcm/dtconv.h
 *
 *  Created on: Dec 13, 2013
 *      Author: tnemeth
 */

#ifndef DTCONV_H_
#define DTCONV_H_

#include <dcmtk/dcmdata/dctk.h>
#include "pbdcm/dtregistry.h"

namespace pbtz {

/**
 * A Date/Time element pair get from the DICOM dataset.
 * Multiple values ( VM > 1 ) are also supported if both the date and time element has the same value multiplicity.
 */
class DateTimeElementPair {

		/** The date element */
		DcmDate * m_date;
		/** The time element */
		DcmTime * m_time;

		/** Date/Time tag pair definition */
		const DateTimeTagPair m_tagPair;

		/** Private block for PB Private elements within the private group */
		Uint16 m_privateGroup;


		/**
		 * Constructor for creating a new DateTimeElements object
		 * This constructor is intended to be used by friend function 'convert(DcmDate &, DcmTime &, const OFString &, const OFString &)' only.
		 *
		 * @param date   The date element
		 * @param date   The time element
		 */
		DateTimeElementPair( DcmDate & date, DcmTime & time ) :
		m_tagPair( date.getTag(), time.getTag() ) {
			m_date = &date;
			m_time = &time;
			m_privateGroup = 0;
		}

		/**
		 * Friend function converting date/time from one UTC offset to another.
		 */
		friend bool convert( DcmDate & date, DcmTime & time, const OFString & fromUTCOffset, const OFString & toUTCOffset );

	public:

		/**
		 * Constructor for creating a new DateTimeElements object
		 *
		 * @param tagPair        defines the DICOM tags of the date and time elements
		 * @param privateGroup   private block reserved for PB Private elements (optional)
		 */
		DateTimeElementPair( const DateTimeTagPair & tagPair, const Uint16 & privateGroup = 0 ) : m_tagPair( tagPair ) {
			m_date = NULL;
			m_time = NULL;
			m_privateGroup = privateGroup;
		}

		/**
		 * Destructor to destroy this DateTimeElements object
		 */
		~DateTimeElementPair() {
		}

		/**
		 * @return true if and only if time part is set in this Date/Time element
		 */
		inline bool hasTime() const {
			return m_time != NULL;
		}

		/**
		 * Sets the date part of this Date/Time element
		 */
		inline void setDate( DcmDate * d ) {
			m_date = d;
		}

		/**
		 * Sets the time part of this Date/Time element
		 */
		inline void setTime( DcmTime * t ) {
			m_time = t;
		}

		/**
		 * @return the date part of this Date/Time element
		 */
		inline DcmDate * getDate() const {
			return m_date;
		}

		/**
		 * @return the time part of this Date/Time element
		 */
		inline DcmTime * getTime() const {
			return m_time;
		}


		/**
		 * @return the DICOM tag of the date part of this Date/Time element
		 */
		inline const DcmTagKey & getDateTag() const {
			return m_tagPair.getDate();
		}

		/**
		 * Checks whether the provided DICOM tag matches to the time tag of this Date/Time element.
		 * In case of private tags the reserved block will also be checked.
		 * @param tag  the tag to check
		 * @return true if and only if <code>tag</code> matches to the time tag of this Date/Time element
		 */
		inline bool isTimeTagMatches( const DcmTag & tag ) {
			const PrivateDateTimeTagPair * pdt = dynamic_cast<const PrivateDateTimeTagPair *>( &m_tagPair );
			if ( pdt != NULL ) {
				return pdt->isTimeTagMatches( tag, m_privateGroup );
			}
			return m_tagPair.isTimeTagMatches( tag );
		}

		/**
		 * Prints this Date/Time element to the standard output for debug purposes
		 */
		void print();

		/**
		 * Calculates Time Zone offset for the time-stamp specified by this Date/Time element according
		 * to the day light savings of the current timezone set.
		 *
		 * @param timeZone       output parameter for the time zone abbreviation
		 * @param offsetSeconds  output parameter for the offset seconds from UTC
		 *
		 * @return true if and only if the time zone offset could be successfully calculated
		 */
		bool calcUTCOffset( OFString & timeZone, long int & offsetSeconds );

		/**
		 * Converts the time-stamp specified by this Date/Time element from one timezone to another.
		 *
		 * @param fromTimeZone        time zone to be converted from (NULL means timezone is already set)
		 * @param toTimeZone          target time zone (will be ignored if fromTimeZone is NULL)
		 * @param toUTCOffsetSeconds  target UTC offset - will not be used if both fromTimeZone and toTimeZone is set
		 *
		 * supported combinations:
		 *                                                fromTimeZone       toTimeZone       toUTCOffsetSeconds
		 * Filling UTC offset for non-local timezone          NULL              NULL                x
		 * DICOM offset to local                               x                 x                 NULL
		 * local to DICOM offset                               x                 x                  x (for optimization only)
		 *
		 * @return true if and only if the time-stamp was successfully converted or no conversion was needed
		 */
		bool convertOffset( const char * fromTimeZone, const char * toTimeZone, const long int * toUTCOffsetSeconds );
};


/**
 * A DateTime (DT) element get from the DICOM dataset.
 * Multiple values ( VM > 1 ) are also supported.
 */
class TimestampElement {
		/** The DateTime element */
		DcmDateTime * m_timestamp;

	public:
		/**
		 * Constructor for creating a new TimestampElement object
		 *
		 * @param timestamp      the DICOM DateTime element
		 */
		TimestampElement( DcmDateTime * timestamp ) {
			m_timestamp = timestamp;
		}

		/**
		 * Prints this DateTime element to the standard output for debug purposes
		 */
		void print();


		/**
		 * Converts the time-stamp specified by this DateTime element from one timezone to another.
		 *
		 * @param fromTimeZone        time zone to be converted from (NULL means timezone is already set)
		 * @param toTimeZone          target time zone (will be ignored if fromTimeZone is NULL)
		 * @param toUTCOffsetSeconds  target UTC offset - will not be used if both fromTimeZone and toTimeZone is set
		 *
		 * supported combinations:
		 *                                                fromTimeZone       toTimeZone       toUTCOffsetSeconds
		 * Filling UTC offset for non-local timezone          NULL              NULL                x
		 * DICOM offset to local                               x                 x                 NULL
		 * local to DICOM offset                               x                 x                  x (for optimization only)
		 *
		 * @return true if and only if the time-stamp was successfully converted or no conversion was needed
		 */
		bool convertOffset( const char * fromTimeZone, const char * toTimeZone, const long int * toUTCOffsetSeconds );
};


/**
 * List of (supported) Date/Time (DA/TM) element pairs in the DICOM object
 */
class DateTimeElementList {
		/** The list of Date/Time element pairs */
		OFList< DateTimeElementPair *> m_list;

	public:

		/**
		 * Constructor for creating a new DateTimeElementList object
		 */
		DateTimeElementList() {
		}

		/**
		 * Constructor to destroy this new DateTimeElementList object
		 */
		~DateTimeElementList();


		/**
		 * Adds the date element to the list.
		 *
		 * A new DateTimeElementPair object using the provided date element will be created and added to the list.
		 * @param dateElem  the date element to be added to the list
		 * @return true if and only if <code>dateElem</code> was successfully added to the list
		 */
		bool addDate( DcmDate * dateElem );

		/**
		 * Adds the time element to the list.
		 *
		 * Finds a matching DateTimeElementPair object for the provided time element in the list and
		 * sets the time part of the DateTimeElementPair object to use the provided time element
		 * @param timeElem  the time element to be added to the list
		 * @return true if and only if <code>timeElem</code> was successfully added to the list
		 */
		bool addTime( DcmTime * timeElem );

		/**
		 * Prints list content to the standard output for debug purposes
		 */
		void print();

		/**
		 * Calculates Time Zone offset for the time-stamps in this list according
		 * to the day light savings of the current timezone set.
		 *
		 * @param timeZone       output parameter for the time zone abbreviation
		 * @param offsetSeconds  output parameter for the offset seconds from UTC
		 *
		 * @return true if and only if the time zone offset could be successfully calculated
		 */
		bool calcUTCOffset( OFString & offset, long int & offsetSeconds );

		/**
		 * Converts the time-stamps for each DateTimeElementPair object in this list from one timezone to another.
		 *
		 * @param fromTimeZone        time zone to be converted from (NULL means timezone is already set)
		 * @param toTimeZone          target time zone (will be ignored if fromTimeZone is NULL)
		 * @param toUTCOffsetSeconds  target UTC offset - will not be used if both fromTimeZone and toTimeZone is set
		 *
		 * supported combinations:
		 *                                                fromTimeZone       toTimeZone       toUTCOffsetSeconds
		 * Filling UTC offset for non-local timezone          NULL              NULL                x
		 * DICOM offset to local                               x                 x                 NULL
		 * local to DICOM offset                               x                 x                  x (for optimization only)
		 *
		 * @return true if and only if all elements in this list was successfully converted or no conversion was needed
		 */
		bool convertOffsets( const char * fromTimeZone, const char * toTimeZone, const long int * toUTCOffsetSeconds );

		/**
		 * Removes all elements from the list.
		 */
		void clear();
};


/**
 * List of (supported) DateTime (DT) elements in the DICOM object
 */
class TimestampList {
		/** The list of DT elements */
		OFList< TimestampElement *> m_list;

	public:

		/**
		 * Constructor creating a new TimestampList object
		 */
		TimestampList() {
		}

		/**
		 * Destructor to destroy this TimestampList object
		 */
		~TimestampList();

		/**
		 * Adds a DateTime (DT) element to this list
		 * @param dateTimeElem  the DateTime element to add
		 * @return true
		 */
		inline bool addDateTime( DcmDateTime * dateTimeElem ) {
			m_list.push_back( new TimestampElement( dateTimeElem ) );
			return true;
		}

		/**
		 * Prints list content to the standard output for debug purposes
		 */
		void print();

		/**
		 * Converts for each TimestampElement object in this list from one timezone to another.
		 *
		 * @param fromTimeZone        time zone to be converted from (NULL means timezone is already set)
		 * @param toTimeZone          target time zone (will be ignored if fromTimeZone is NULL)
		 * @param toUTCOffsetSeconds  target UTC offset - will not be used if both fromTimeZone and toTimeZone is set
		 *
		 * supported combinations:
		 *                                                fromTimeZone       toTimeZone       toUTCOffsetSeconds
		 * Filling UTC offset for non-local timezone          NULL              NULL                x
		 * DICOM offset to local                               x                 x                 NULL
		 * local to DICOM offset                               x                 x                  x (for optimization only)
		 *
		 * @return true if and only if all elements in this list was successfully converted or no conversion was needed
		 */
		bool convertOffsets( const char * fromTimeZone, const char * toTimeZone, const long int * toUTCOffsetSeconds );

		/**
		 * Removes all elements from the list.
		 */
		void clear();
};


/**
 * Specifies modes for how to proceed in collectTimes according to the existence or non-existence of a (0008,0201) TimezoneOffsetFromUTC tag
 */
typedef enum {
	/** Stop processing if the TimezoneOffsetFromUTC tag exists */
	UTCTAG_STOP_IF_EXISTS = 0,
	/** Stop processing if the TimezoneOffsetFromUTC tag does not exist */
	UTCTAG_STOP_IF_DOESNT_EXIST = 1,
	/** Proceed regardless to the existence or non-existence of the TimezoneOffsetFromUTC tag */
	UTCTAG_IGNORE = 2
} OFFSET_TAG_ACTION;


/**
 * Base class for several date/time conversions between time zones
 */
class DateTimeConverter {
	protected:

		/** The dataset */
		DcmDataset * m_dataset;

		/** List of Date/Time (DA/TM) element pairs in the dataset */
		DateTimeElementList m_dateTimeList;
		/** List of DateTime (DT) elements in the dataset */
		TimestampList m_dtstampList;

		/** Offset from UTC set in the dataset - in (0008,0201) TimezoneOffsetFromUTC tag */
		OFString m_datasetUTCOffset;
		/** System timezone set on the local computer */
		OFString m_systemTimeZone;


		/** Will be set to true in collectTimes if dataset contains a non-empty (0008,0201) TimezoneOffsetFromUTC tag */
		bool m_hasOffsetInDataset;
		/** true indicates that no time conversion needed for the dataset */
		bool m_dontConvert;

		/**
		 * Prints times in the given item to standard output for debug purposes
		 * @param item   print times from this item recursively
		 */
		void printTimes( DcmItem * item );

		/**
		 * Collect time elements from this item recursively.
		 * Will fill m_dateTimeList and m_dtstampList.
		 * Will also set m_hasOffsetInDataset and m_dontConvert.
		 * Collecting the time elements regarding to whether a UTC offset is set or not can be controlled using utcOffsetMode in the following way:
		 *
		 * UTCTAG_STOP_IF_EXISTS : stops collecting the time elements if UTC offset is set
		 * UTCTAG_STOP_IF_DOESNT_EXIST : stops collecting the time elements if UTC offset is _not_ set
		 * UTCTAG_IGNORE : collects all time elements
		 *
		 * @param item           collect time elements from this item
		 * @param utcOffsetMode  controls whether to collect all time elements or not regarding to the status of the UTC offset tag
		 *
		 * @return true if and only if time elements were successfully collected
		 */
		bool collectTimes( DcmItem * item, OFFSET_TAG_ACTION utcOffsetMode );


	public:

		/**
		 * Constructor creating a new DateTimeConverter object
		 * @param dataset  the dataset of the DICOM object to convert
		 */
		DateTimeConverter( DcmDataset * dataset );
		/**
		 * Destructor to destroy this DateTimeConverter object
		 */
		virtual ~DateTimeConverter();

		/**
		 * Prints times in the dataset to the standard output for debug purposes
		 */
		void printTimes();
};

/**
 * Converter to convert times in a DICOM object to local timezone.
 * If no (0008,0201) TimezoneOffsetFromUTC is set in the DICOM object, then the times will not be converted.
 * If (0008,0201) TimezoneOffsetFromUTC is set, then all times in the DICOM objects will be converted to the
 * local timezone according to the actual daylight saving mode at each time stamp.
 */
class Dicom2LocalConverter : public DateTimeConverter {

	public:

		/**
		 * Constructor creating a new Dicom2LocalConverter object
		 * @param dataset  the dataset of the DICOM object to convert
		 */
		Dicom2LocalConverter( DcmDataset * dataset );
		/**
		 * Destructor to destroy this Dicom2LocalConverter object
		 */
		virtual ~Dicom2LocalConverter();

		/**
		 * Converts all times in the DICOM object to local time
		 * @return true if and only if all timestamps were successfully converted to local time
		 */
		bool convert();
};


/**
 * Adjust date/times in a dataset to a given offset.
 *
 * adjust( const OFString & ) can be called several times without recollecting times and offset information over and over
 */
class OffsetAdjustedDataset : public DateTimeConverter {
		/** true indicates that collectTimes() was already run */
		bool m_timesCollected;

	public:
		/**
		 * Constructor creating a new OffsetConverter object
		 * @param dataset  the dataset of the DICOM object to convert
		 */
		OffsetAdjustedDataset( DcmDataset * dataset );
		/**
		 * Destructor to destroy this OffsetConverter object
		 */
		virtual ~OffsetAdjustedDataset();

		/**
		 * Converts all timestamps in the DICOM dataset to the given offset from UTC.
		 * (0008,0201) TimezoneOffsetFromUTC will be updated in the dataset
		 *
		 * targetOffset intended to be in the format defined and accepted by the DICOM standard.
		 * Empty offset means local time (according to actual daylight saving of particular timestamps)
		 *
		 * @param targetOffset   convert timestamps to this offset from UTC
		 *
		 * @return true if and only if all timestamps were successfully converted
		 */
		bool adjust( const OFString & targetOffset );


		/**
		 * Converts all timestamps in this DICOM dataset according to the offset from UTC set in the parameter dataset <code>dset</code>.
		 * (0008,0201) TimezoneOffsetFromUTC will be updated in this dataset (the parameter dataset <code>dset</code> shall not be affected)
		 *
		 * @param dset   adjust offset to be the same as in <code>dset</code>.
		 *
		 * @return true if and only if all timestamps were successfully converted
		 */
		bool adjust( /*const - DCMTK does not define findAndGet... methods as const */ DcmDataset & dset );

		inline DcmDataset * getDataset() {
			return m_dataset;
		}

		/**
		 * Clears the Date (DA)/Time (TM) and Datetime (DT) tag lists to force the actual tags to be re-collected.
		 * Intended to be called when elements got removed from the underlying dataset since last recollect.
		 */
		inline void clearLists() {
			m_dateTimeList.clear();
			m_dtstampList.clear();
			m_timesCollected = false;
		}

};


/**
 * Converts date/times in a dataset to a given timezone.
 *
 * Unlike OffsetAdjustedDataset, DicomToTimeZone converts to a given timezone and leaves TimezoneOffsetFromUTC untouched.
 * According to this, multiple conversions are not supported.
 *
 * Datasets converted using DicomToTimezone are not intended to be saved.
 */
class DicomToTimeZone : public DateTimeConverter {
		/** true indicates that convert() was already called */
		bool m_timesConverted;

	public:
		/**
		 * Constructor creating a new DicomToTimeZone object
		 * @param dataset  the dataset of the DICOM object to convert
		 */
		DicomToTimeZone( DcmDataset * dataset );

		/**
		 * Destructor to destroy this DicomToTimeZone object
		 */
		virtual ~DicomToTimeZone();

		/**
		 * Converts all timestamps in the DICOM dataset to the given timezone.
		 *
		 * Empty timezone means local time (according to actual daylight saving of particular timestamps)
		 *
		 * @param timezone   convert timestamps to this timezone
		 *
		 * @return true if and only if all timestamps were successfully converted
		 */
		bool convert( const OFString & timezone );

		inline DcmDataset * getDataset() {
			return m_dataset;
		}
};


/**
 * Converts date/times in a dataset from a given timezone to local timezone.
 *
 * If the dataset contains TimezoneOffsetFromUTC, then the date/time values in the dataset remain untouched.
 * If the dataset contains no TimezoneOffsetFromUTC, then the date/time values will be converted to local time.
 * TimezoneOffsetFromUTC shall not be added to the dataset.
 *
 * Intended to be used to convert date/times mapped from HL7 messages
 */
class TimeZoneToLocal : public DateTimeConverter {

	public:
		/**
		 * Constructor creating a new TimeZoneToLocal object
		 * @param dataset  the dataset to convert
		 */
		TimeZoneToLocal( DcmDataset * dataset );

		/**
		 * Destructor to destroy this TimeZoneToLocal object
		 */
		virtual ~TimeZoneToLocal();

		/**
		 * Converts all timestamps in the DICOM dataset from the given timezone to local time zone.
		 *
		 * Empty timezone means local time (according to actual daylight saving of particular timestamps)
		 *
		 * @param timezone   convert timestamps from this timezone
		 *
		 * @return true if and only if all timestamps were successfully converted
		 */
		bool convert( const OFString & timezone );
};


/**
 * Converter to convert a DICOM dataset having no UTC offset defined to the specified timezone.
 * The UTC offset will be set according to the timezone. The time elements will also be adjusted
 * according to the dayligth savings within the timezone.
 *
 * The dataset will not be touched if (0008,0201) TimezoneOffsetFromUTC is already set.
 */
class TimeZoneFiller : public DateTimeConverter /* : public Local2DicomConverter */ {

		/** Name of the timezone to convert to */
		OFString m_tzname;

		/** String representation of the UTC offset to be set for the object */
		OFString m_timeZone;
		/** Offset from UTC in seconds to be set for the object */
		long int m_offset;

		/** Adjust time elements according to the UTC offset and daylight saving times in the timezone */
		bool adjustTimes();

	public:
		/**
		 * Constructor creating a new TimeZoneFiller object
		 * @param dataset  the dataset to be converted
		 * @param tzname   name of the target timezone
		 */
		TimeZoneFiller( DcmDataset * dataset, const char * tzname );
		/**
		 * Destructor to destroy this TimeZoneFiller object
		 */
		virtual ~TimeZoneFiller();

		/**
		 * Calculates and fills offset from UTC for the dataset. Also adjust the time elements if needed.
		 *
		 * @return true if and only if the offset from UTC was successfully set and time elements successfully adjusted in the dataset
		 */
		bool fillTimeZone();

		/**
		 * Returns the DICOM UTC offset defined in the dataset after the conversion
		 * Intended to be called after fillTimeZone() only.
		 *
		 * @return  the DICOM UTC offset defined in the dataset after the conversion
		 */
		inline const OFString & getUTCOffset() const {
			return m_hasOffsetInDataset ? m_datasetUTCOffset : m_timeZone;
		}
};


/*
 * Returns current time converted to a given UTC offset. UTC offset intended to be defined in DICOM supported format.
 *
 * @param utcOffset   timezone offset from UTC
 * @param outDate     output parameter for the date part
 * @param outTime     output parameter for the time part
void nowInUTCOffset( const OFString & utcOffset, OFString & outDate, OFString & outTime );
 */


/**
 * Returns current time converted to a given UTC offset. UTC offset intended to be defined in DICOM supported format.
 *
 * @param utcOffset   timezone offset from UTC
 * @param outNow      output parameter for current time
 */
void nowInUTCOffset( const OFString & utcOffset, struct tm & outNow );

/**
 * Converts a date/time stored in separate Date and Time elements from one time zone to another.
 * Time zones are intended to be offsets from UTC as defined by the DICOM standard.
 * Empty offset means offset at the specified time interpreted in the local time zone (according to daylight savings).
 *
 * @param date        date element holding the date part of the date/time to convert
 * @param time        time element holding the time part of the date/time to convert
 * @param fromOffset  current offset of the date/time
 * @param toOffset    target offset
 *
 * @return true if and only if the date/time was successfully converted to the specified offset
 */
bool convert( DcmDate & date, DcmTime & time, const OFString & fromOffset, const OFString & toOffset );

}

#endif /* DTCONV_H_ */
