/*
 *  Module:  pb-scp
 *
 *  Header file for mapping dicom tags.
 *
 */

#include <map>
#include <string.h>

#include <dcmtk/ofstd/ofstring.h>
#include <dcmtk/ofstd/ofuoset.h>
#include <dcmtk/dcmdata/dctk.h>

#include "pbdcm/PBCustomFields.h"


#define DEBUGLOG(...) {openlog("PBMap", LOG_PID, LOG_LOCAL3); syslog(LOG_DEBUG, __VA_ARGS__); closelog(); }


/**
 * Collection of data sources (dataset, temporary variables, ...) used during a single mapping session.
 *
 * Special / Control variables supported during a mapping session:
 *
 * @PROCESS - whether the object need to be processed after the coercion (default: true)
 *
 */
class MappingSession {
	private:
		/** Dataset of the DICOM object the mapping being applied to */
		DcmDataset & m_dataset;

		/** Custom Defined fields in the DICOM object */
		PBCustomFields m_customFields;

		/** Temporary variables created during mapping session */
		std::map<OFString, OFString> m_tempVars;

		/** Special/control variables */
		std::map<std::string, std::string> m_specials;


	public:

		/**
		 * Constructor for creating a new Mapping Session to apply mappings to a DICOM object.
		 *
		 * @param dataset   Dataset of the DICOM objects the mappings to be applied to.
		 */
		MappingSession( DcmDataset & dataset ) :m_dataset( dataset ), m_customFields( dataset ) {
			m_specials[ "PROCESS" ] = "true";
		}

		/**
		 * Gets dataset of the DICOM object associated with this Mapping Session.
		 *
		 * @return  Dataset of the DICOM obect to be manipulated.
		 */
		inline DcmDataset & getDataset() const {
			return m_dataset;
		}

		/**
		 * Returns value for custom defined user field.
		 *
		 * @param name   Name of the custom defined field
		 *
		 * @return  Value of the custom defined field or NULL if the custom defined field is not yet set
		 */
		inline const char * getCustomField( const char * fieldName ) const {
			return m_customFields.getFieldValue( fieldName );
		}

		/**
		 * Sets value for a custom defined user field.
		 *
		 * @param name    Name of the custom defined field
		 * @param value   New value for the custom defined field (may be NULL to clear the field)
		 *
		 * @return true if and only if the new value was successfully set for the field
		 */
		inline bool setCustomField( const char * fieldName, const char * newValue ) {
			return m_customFields.setFieldValue( fieldName, newValue );
		}

		/**
		 * Returns value for temporary variable.
		 *
		 * @param name   Name of the temporary variable
		 *
		 * @return  Value of the temporary variable or NULL if the temporary variable is not yet set
		 */
		const char* const getVariable( const char* const name ) const;

		/**
		 * Sets value for a temporary variable.
		 *
		 * @param name    Name of the temporary variable
		 * @param value   New value for the temporary variable (may be NULL)
		 */
		void setVariable( const char* const name, const char* const value );

		/**
		 * Clears the temporary variables
		 */
		inline void clearVariables() {
			m_tempVars.clear();
		}

		/**
		 * Sets the PROCESS special/control variable
		 * @param process new value for the PROCES control variable
		 */
		inline void setProcess( bool process ) {
			if ( process ) {
				m_specials[ "PROCESS" ] = "true";
			}
			else {
				m_specials.erase( "PROCESS" );
			}
		}

		/**
		 * Gets the PROCESS special/control variable
		 * @return the value of the PROCES control variable
		 */
		inline bool isProcess() const {
			try {
				m_specials.at( "PROCESS" );
				return true;
			}
			catch ( const std::out_of_range& e ) {
				return false;
			}
		}
};


class Value;


/**
 * Base class for possible lvalue objects.
 * Supported lvalue objects:
 *    - dicom tag
 *    - custom defined field
 *    - temporary variable
 */
class LVObject {
	protected:
		/** Constructor for creating a left value object */
		LVObject() {
		}

	public:

		/** Destructor to destroy the left value object */
		virtual ~LVObject();

		/**
		 * Assign value to the element represented by this left value object in mapping session.
		 *
		 * @param msession  mapping session
		 * @param value     value to be assigned to this lv object
		 */
		virtual void assignValue( MappingSession & msession, Value * value ) = 0;
};

/**
 * Left value object to calculate/store value for DICOM tags.
 */
class LVDicomTag : public LVObject {
	protected:

		/** Tag Key of the DICOM tag the calculated value to be stored to */
		DcmTag m_targetKey;

		/** Constructor for creating a left value object */
		LVDicomTag() {
		}

	public:
		/** Constructor for creating a left value object to store values for a DICOM tag
		 *
		 * @param targetKey  Tag Key of the DICOM tag the calculated value to be stored to
		 */
		LVDicomTag( const DcmTag & targetKey ) {
			m_targetKey = targetKey;
		}

		/**
		 * Copy Constructor.
		 *
		 * @param orig  Original LVDicomTag object to be copied
		 */
		LVDicomTag( const LVDicomTag & orig ) {
			m_targetKey = orig.m_targetKey;
		}

		/**
		 * @return Tag Key of the DICOM tag represented by this left value object
		 */
		inline const DcmTag & getTargetTag() const {
			return m_targetKey;
		}

		/**
		 * operator=
		 * @param orig   LVDicomTag object values to be copied from
		 * @return
		 */
		LVDicomTag & operator=( const LVDicomTag & orig ) {
			m_targetKey = orig.m_targetKey;
			return *this;
		}

		/** Destructor to destroy the LV Dicom Tag object */
		virtual ~LVDicomTag();

		/**
		 * Assign value to the DICOM tag represented by this LV DICOM Tag object in the dataset of the mapping session.
		 *
		 * @param msession  mapping session
		 * @param value     value to be assigned to this lv object
		 */
		virtual void assignValue( MappingSession & msession, Value * value );

		void assignValueInner( DcmItem & parentItem, Value * value );

};


/**
 * Left value object to calculate/store value for DICOM tags in a Sequence
 */
class LVSeqDicomTag : public LVDicomTag {
	private:

		/**
		 * The parameters (gtag, etag, itemno ...) are stored here
		 */
		OFList< int > m_args;

		/**
		 * operator=
		 * @param orig   LVDicomTag object values to be copied from
		 * @return
		 */
		LVSeqDicomTag & operator=( const LVSeqDicomTag & orig ) {
			m_targetKey = orig.m_targetKey;
			return *this;
		}

	public:
		/**
		 * Constructor for creating a left value object to store values for a DICOM tag
		 *
		 * @param targetKey  Tag Key of the DICOM tag the calculated value to be stored to
		 * @param args			The parameters of the Sequence Dicom tag
		 */
		LVSeqDicomTag( const DcmTag & targetKey, const OFList< int > & args );

		/**
		 * Copy Constructor.
		 *
		 * @param orig  Original LVDicomTag object to be copied
		 */
		LVSeqDicomTag( const LVSeqDicomTag & orig );

		/**
		 * @return Tag Key of the DICOM tag represented by this left value object
		 */
		inline const DcmTag & getTargetTag() const {
			return m_targetKey;
		}



		/** Destructor to destroy the LV Dicom Tag object */
		virtual ~LVSeqDicomTag();

		/**
		 * Assign value to the DICOM tag represented by this LV DICOM Tag object in the dataset of the mapping session.
		 *
		 * @param msession  mapping session
		 * @param value     value to be assigned to this lv object
		 */
		virtual void assignValue( MappingSession & msession, Value * value );
};

/**
 * Left value object to calculate/store value for Custom Defined fields.
 */
class LVCustomField : public LVObject {
	private:

		/** Name of the Custom Defined field */
		OFString m_fieldName;

	public:

		/** Constructor for creating a left value object to store values for a custom defined field
		 *
		 * @param fieldName   Name of the Custom Defined field
		 * @return
		 */
		LVCustomField( const OFString & fieldName ) {
			m_fieldName = fieldName;
		}

		/**
		 * Copy Constructor.
		 *
		 * @param orig  Original LVDicomTag object to be copied
		 */
		LVCustomField( const LVCustomField & orig ) {
			m_fieldName = orig.m_fieldName;
		}

		/**
		 * @return Name of the User Defined Custom field represented by this left value object
		 */
		const char * getFieldName() const {
			return m_fieldName.empty() ? NULL : m_fieldName.c_str();
		}

		/**
		 * operator=
		 * @param orig   LVDicomTag object values to be copied from
		 * @return
		 */
		LVCustomField & operator=( const LVCustomField & orig ) {
			m_fieldName = orig.m_fieldName;
			return *this;
		}

		/** Destructor to destroy the LV Custom Field object */
		virtual ~LVCustomField();


		/**
		 * Assign value to the Custom Defined field represented by this left value object in the dataset of the mapping session.
		 *
		 * @param msession  mapping session
		 * @param value     value to be assigned to this lv object
		 */
		virtual void assignValue( MappingSession & msession, Value * value );
};

/**
 * Left value object to calculate/store value for temporary variables.
 */
class LVTempVar : public LVObject {
	private:

		/** Name of the temporary variable */
		OFString m_varName;

	public:

		/** Constructor for creating a left value object to store values to a temporary variable
		 *
		 * @param varName   Name of the temporary variable
		 * @return
		 */
		LVTempVar( const OFString & varName ) {
			m_varName = varName;
		}

		/**
		 * Copy Constructor.
		 *
		 * @param orig  Original LVDicomTag object to be copied
		 */
		LVTempVar( const LVTempVar & orig ) {
			m_varName = orig.m_varName;
		}

		/**
		 * operator=
		 * @param orig   LVDicomTag object values to be copied from
		 * @return
		 */
		LVTempVar & operator=( const LVTempVar & orig ) {
			m_varName = orig.m_varName;
			return *this;
		}

		/** Destructor to destroy the LV Temp Var object */
		virtual ~LVTempVar();

		/**
		 * Assign value to the temporary variable in the mapping session.
		 *
		 * @param msession  mapping session
		 * @param value     value to be assigned to this lv object
		 */
		virtual void assignValue( MappingSession & msession, Value * value );
};



/**
 * Value class
 *
 *
 */

class Value {
	public:

		Value() {
			m_refCount = 1;
			m_value = NULL;
			m_bStringDone = false;
		}

	protected:

		virtual ~Value();

		const char * setString( const char * s ) {
			if ( m_value != NULL )
				delete[] m_value;

			if ( s != NULL ) {
				m_value = new char[ strlen( s ) + 2 ];
				strcpy( m_value, s );
			}
			else {
				m_value = new char[ 2 ];
				m_value[ 0 ] = '\0';
			}
			return m_value;
		}

	public:

		static Value * ref( Value * value ) {
			if ( value != NULL ) {
				value->m_refCount++;
			}
			return value;
		}

		static void deref( Value * value ) {
			if ( value != NULL && --(value->m_refCount) == 0 ) {
				delete value;
			}
		}

		// functions to overload

	protected:

		virtual void makeString() {
		};

	public:

		// assigns this value to te (target element)
		virtual void assignTo( DcmItem * di, DcmElement * te ) const = 0;

		// returns the value as a string
		virtual const char * getString() {
			if ( !m_bStringDone ) {
				makeString();
				m_bStringDone = true;
			}

			return m_value;
		}

		// returns the value as an element reference
		virtual operator DcmElement * () const {
			return NULL;
		}

	private:

		bool m_bStringDone;
		int m_refCount;

	protected:
		char * m_value;
};

// tag reference value (such as a (xxxx,yyyy) )

class TagValue : public Value {
	public:

		TagValue( DcmElement * elem ) : Value() {
			m_elem = elem;
		}

	protected:

		void makeString() {
			if ( m_elem == NULL )
				return;

			char * vv = NULL;
			m_elem->getString( vv );
			setString( vv );
		}

		virtual operator DcmElement * () const {
			return m_elem;
		}

		virtual void assignTo( DcmItem * di, DcmElement * te ) const;

	private:
		DcmElement * m_elem;
};

// string value;

class StringValue: public Value {
	public:
		StringValue( const char * s ) : Value() {
			setString( s );
		}

		virtual void assignTo( DcmItem * di, DcmElement * te ) const;
};


class Expression;


/**
 * Function class
 *
 *
 */

class Func {

	public:

		Func( const char * name, const char * opname, Value * (*getValue) ( MappingSession & msession, OFList< Expression * > & args ), bool ( *numParamsOK )( int n ) ) {
			m_name = name;
			m_opname = opname;
			m_getValue = getValue;
			m_numParamsOK = numParamsOK;
		}

		Value * getValue( MappingSession & msession, OFList< Expression * > & args ) {
			return m_getValue( msession, args );
		}

		bool numParamsOK( int n ) {
			return m_numParamsOK( n );
		}

		const char * getName() {
			return m_name;
		}

		const char * getOpName() {
			return m_opname;
		}

	private:
		const char * m_name;

		/** Certain functions (operations) can have a short name (like 'add' -> '+')
		 * [Why the fuck this can't be documented by the author]
		 */
		const char * m_opname;
		Value * ( *m_getValue ) ( MappingSession & msession, OFList< Expression * > & args );
		bool ( *m_numParamsOK )( int n );
};


/**
 * Subclass for implementing the =func(...) expressions
 *
 */

class Expression {
	public:

		// func such as func(... , ... , ...) or ... [op] ...
		Expression(Func * f = NULL) {
			m_func = f;
			m_value = NULL;
		}

		// constant literal value
		Expression( char * value ) {
			m_func = NULL;
			m_value = value;
		}

		~Expression();

		static Expression * readCfgLine( const char * line );

		Value * getValue( MappingSession & msession );

		void print();

		friend std::ostream& operator<<( std::ostream & os, const Expression & expr );

	protected:
		// advances the parser's char ptr with skipping quotes etc.
		char * parserStep( char * p );

		// creates args (splits up by commas)
		// returns error string
		char * parseParamList( char * line );

		// parses an expression
		// returns error string
		char * parse( char *line );

		char * m_value;
		Func * m_func;
		OFList< Expression * > m_args;
};


/**
 * Class that implements a single assigment line in a mapping file
 */
class Assignment {
	public:

		Assignment() {
			m_lvObject = NULL;
			m_expression = NULL;
		};

		~Assignment() {
			if ( m_expression != NULL ) {
				delete m_expression;
				m_expression = NULL;
			}
			if ( m_lvObject != NULL ) {
				delete m_lvObject;
				m_lvObject = NULL;
			}
		};

		/**
		 * Returns and detaches pointer to the left value object of the assignment.
		 * Caller should make sure the pointer will be freed.
		 *
		 * @return pointer to the Left Value Object of the assignment
		 */
		inline LVObject * getAndDetachLVObject() {
			LVObject * rv = m_lvObject;
			m_lvObject = NULL;
			return rv;
		}

		bool readCfgLine( const char * line );
		int apply( MappingSession & msession );

	private:
		LVObject * m_lvObject;
		Expression * m_expression;

};

/**
 * @enum GlobalMappingType
 *
 * Supported global coercion rules
 */
enum class GlobalMappingRule {
	/// Global preceding coercion rule
	PRE,
	/// Global post coercion rule
	POST
};


/**
 * Class that implements a whole mapping file (or rules in string)
 *
 */

class Mapping {
	public:

		Mapping(FILE * f, const char * n);

		/**
		 * Constructor that creates mapping from the rules string
		 *
		 * @param rules	the coercion rules
		 * @param n	the name of the mapping
		 */
		Mapping(const char * rules, const char * n);
		~Mapping();

		int apply( MappingSession& ms );

		/**
		 * Reads mapping rules from rule file and applies to the provided DICOM object.
		 * If lvList is provided, pointer to Left Value Objects (dicom tags, custom defined user fields) affected by the
		 * mapping rules will be added to this list.
		 * The pointers added to the list should be freed by the caller.
		 *
		 * @param fileName   Name of the mapping rules file
		 * @param ds         Dataset of the DICOM object the rules to be applied to
		 * @param lvList     Output parameter for the list of modified tags/user fields
		 *
		 * @return  true if and only if the rules could be applied to the dataset successfully
		 */
		static bool applyStatic( const char* fileName, MappingSession& ms, OFList<LVObject*>* lvList = NULL );

		inline const char * getName() const {
			return m_name;
		};

	private:
		OFList < Assignment * > m_assignments;
		char m_name[128];
};


/**
 * Class that implements a set of mapping files each corresponding to an AETITLE
 * It is wise to instanciate one copy of MappingSet on startup (and reinstanciate
 * on HUP), so it searches/reads/parses the mapping files once on startup.
 * Then, one can call apply many times.
 */

class MappingSet {
	public:
		/**
		 * Creates a new mapping set to be applied to incoming objects.
		 * Reads global inbound rules, and inbound coercion rules for all devices.
		 */
		MappingSet();

		/**
		 * Creates a mapping set for applying coercion when receiving from/sending to a particular device.
		 *
		 * @param aeTitle the AE title of the device (NULL indicates no device - for example media creation)
		 * @param isOutbound true indicates to create the mapping set for outbound coercion
		 * @param readGlobalRules true indicates to also read global coercion rules
		 */
		MappingSet( bool isOutbound, const char *aeTitle, bool readGlobalRules );

		~MappingSet();

		/**
		 * Applies a device (AE title) specific coercion rule to a mapping session (dataset)
		 * @param AEtitle the AE title of the device
		 * @param ms the mapping session
		 * @return 0 on success
		 */
		int apply( const char* AEtitle, MappingSession& ms );

		/**
		 * Applies a global coercion rule to a mapping session (dataset).
		 *
		 * @param globalRule  identifies the global coercion rule to apply
		 * @param ms the mapping session
		 * @return 0 on success
		 */
		int applyGlobal( GlobalMappingRule globalRule, MappingSession &ms );

	private:
		Mapping* readGlobal( const std::string &mappingsDir, const std::string &name, bool isOutbound, std::string &tempName );

		Mapping * m_preGlobal;
		Mapping * m_postGlobal;
		OFList < Mapping * > m_mappings;
};


char * checkSyntaxInner( const char * rules );

/**
 * Helper function for SEQ lvalue and function
 * Gets the parent item from the dataset, defined by args, or NULL, if the item not exists
 *
 * @param	args	in the form of (gtag, etag, itemno, ...) - only the 3x number of args will be used
 * @param	dataset	the dataset to start from
 * @return	the parent item from the dataset, defined by args, or NULL, if the item not exists
 */
DcmItem * getParentItem( OFList< Value * > & args, DcmDataset & ds );

/**
 * Helper function for SEQ lvalue and function
 * Gets the parent item starting from the given item, defined by args, or NULL, if the item not exists
 *
 * @param	args	in the form of (gtag, etag, itemno, ...) - only the 3x number of args will be used
 * @param	orig	the item to start from
 * @return	the parent item from the dataset, defined by args, or NULL, if the item not exists
 */
DcmItem * getParentItem( OFList< int > & args, DcmItem * orig );

