#ifndef UTILS_INCLUDE_UTILS_LOGGER_H_
#define UTILS_INCLUDE_UTILS_LOGGER_H_


#include <cstdio>
#include "imutils/IMLogger.h"

namespace erad::utils {

/**
 * Abstract base class to log messages to various output.
 *
 * One logger object logs at a particular level and to the particular output.
 * Might be useful when different levels log into different output (eg log errors to syslog while log debugs to stdout).
 * A workaround until we have log4c
 */
class Logger {
	protected:

		/**
		 * Constructs a new Logger object.
		 */
		Logger();

	public:

		/**
		 * Destroys the Logger object.
		 */
		virtual ~Logger();

		/**
		 * Logs a formatted message.
		 *
		 * @param format the format string of the log message
		 * @param ... arguments for the format string
		 */
		virtual void logV( const char *format, ... ) = 0;

		/**
		 * Logs a simple message.
		 *
		 * @param msg the message to be logged
		 */
		virtual void log0( const char *msg ) = 0;
};


/**
 * Logger class to log to standard output/error or a file
 */
class StdLogger : public Logger {
	private:
		/// the file handle to log to
		FILE *out;

	public:
		/**
		 * Constructs a new StdLogger object.
		 * The caller should open/close the file handle.
		 *
		 * @param out the file handle to write to
		 */
		StdLogger( FILE *out );

		/**
		 * Destroys the StdLogger object.
		 * Does nothing, the file handle should be closed by the caller.
		 */
		virtual ~StdLogger();

		// Copy/Move constructors and assignment operators are not allowed

		StdLogger( const StdLogger& ) = delete;
		StdLogger( StdLogger&& ) = delete;

		StdLogger& operator=( const StdLogger& ) = delete;
		StdLogger& operator=( StdLogger&& ) = delete;

		/**
		 * @copydoc Logger::logV( const char *, ... )
		 */
		virtual void logV( const char *format, ... ) override;

		/**
		 * @copydoc Logger::log0( const char * )
		 */
		virtual void log0( const char *msg ) override;
};


/**
 * Logger class to log via syslog
 */
class SyslogLogger : public Logger, IMLogger {
	private:
		/// Log level
		LOG_MODE level;

	public:

		/**
		 * Constructs a new SyslogLogger object.
		 *
		 * @param level the log level
		 */
		SyslogLogger( LOG_MODE level );

		/**
		 * Destroys the SyslogLogger object.
		 */
		virtual ~SyslogLogger();

		// Copy/Move constructors and assignment operators are not allowed

		SyslogLogger( const SyslogLogger& ) = delete;

		SyslogLogger( SyslogLogger&& ) = delete;

		SyslogLogger& operator=( const SyslogLogger& ) = delete;
		SyslogLogger& operator=( SyslogLogger&& ) = delete;

		/**
		 * @copydoc Logger::logV( const char *, ... )
		 */
		virtual void logV( const char *format, ... ) override;

		/**
		 * @copydoc Logger::log0( const char * )
		 */
		virtual void log0( const char *msg ) override;
};

/**
 * Logger to discard log messages
 */
class NullLogger : public Logger {
	public:
		/**
		 * Constructs a new NullLogger object to discard log messages.
		 */
		NullLogger();

		/**
		 * Destroys the NullLogger object.
		 */
		virtual ~NullLogger();

		// Copy/Move constructors and assignment operators are not allowed

		NullLogger( const NullLogger& ) = delete;

		NullLogger( NullLogger&& ) = delete;

		NullLogger& operator=( const NullLogger& ) = delete;
		NullLogger& operator=( NullLogger&& ) = delete;

		/**
		 * @copydoc Logger::logV( const char *, ... )
		 */
		virtual void logV( const char *format, ... ) override;

		/**
		 * @copydoc Logger::log0( const char * )
		 */
		virtual void log0( const char *msg ) override;
};

} // namespace

#endif /* UTILS_INCLUDE_UTILS_LOGGER_H_ */
