#ifndef OBJCACHEMGR_H
#define OBJCACHEMGR_H

#include <string>
#include <set>
#include <list>

#include <imutils/DBConnector.h>
#include <imutils/IMLogger.h>

namespace erad::db {

/**
 * Exceptions indicating that the object table entries of a study are not mapped
 * and/or could not be mapped to the object table.
 */
class ObjCacheException : public std::exception {

		/// the error message
		std::string message;

		/// indicates whether error references need to be added to the study/objects
		bool addErrorReferences;

	public:

		/**
		 * Constructs a new Object Cache Exception with a particular error message.
		 * Also logs the message to the error or the debug log according to the <code>logDebugOnly</code> parameter.
		 *
		 * @param message the error message
		 * @param logger for error/debug logging
		 * @param logDebugOnly if <code>true</code>, then logs the message to the debug log
		 * @param addErrorReferences indicates whether `ocerror` references need to be added the the study and/or object files
		 */
		ObjCacheException( const std::string& message, IMLogger* logger, bool logDebugOnly, bool addErrorReferences ) noexcept;

		/**
		 * Constructs a new Object Cache Exception with a particular error message.
		 * Also logs the message to the error log.
		 *
		 * @param message the error message
		 * @param addErrorReferences indicates whether `ocerror` references need to be added the the study and/or object files
		 */
		ObjCacheException( const std::string& message, bool addErrorReferences ) noexcept
				: ObjCacheException( message, NULL, false, addErrorReferences ) {
		}

		/* Use the default copy and move constructors */

		ObjCacheException( const ObjCacheException& orig ) noexcept = default;
		ObjCacheException( ObjCacheException&& orig ) noexcept = default;

		virtual ~ObjCacheException();

		/* Use the default copy and move assignments */

		ObjCacheException& operator=( const ObjCacheException& ) noexcept = default;
		ObjCacheException& operator=( ObjCacheException&& ) noexcept = default;

		/**
		 * @copydoc std::exception::what()
		 */
		virtual const char* what() const noexcept override {
			return message.c_str();
		}

		/**
		 * Indicates whether `ocerror` references need to be added to the study and/or object files.
		 *
		 * If the return value is `true`, then the references need to be added.
		 *
		 * If the return value is `false`, then the references may need not be added, the exception handler may perform
		 * additional checks (eg whether the references actually exist)
		 *
		 * @return `true` if error references need to be added to the study/objects
		 */
		inline bool needsErrorReferences() const {
			return addErrorReferences;
		}
};



class ObjCacheMgr {

	public:

		ObjCacheMgr( erad::db::DBConnector* pDB );

		/**
		 * Checks whether the object table entries of a particular study are mapped to the database.
		 * Maps the object table entries if needed.
		 *
		 * Also touches the study in the object cache table regardless to whether the object entries were already mapped or not.
		 *
		 * When a new order gets created or a study is just received, the object meta file may not yet exist, but we still need to create
		 * an object cache entry.
		 * However a missing object meta file is also a valid  when a study is being deleted,
		 * in which case we do not want to recreate the object cache entry.
		 * Whether an object cache entry to be created when there is no object meta file is controlled by the <code>mapMissing</code> parameter.
		 *
		 * @param study the study instance UID
		 * @param mapMissing optional if <code>true</code> (the default value), then an object cache entry will be created even if no object meta file exists.
		 * @throw IMTimeoutException if could not get lock to import the object table entries
		 * @throw ObjCacheException if an error occurs
		 */
		bool touchObjCache( const std::string& study, bool mapMissing = true );

	protected:

		/**
		 * Performs the actual import by reading the objectdb.meta file and inserting to the Dcobject table
		 *
		 * @param study the study instance UID
		 * @param file path to the objectdb.meta file
		 * @param missingMeta output parameter to indicate that the object meta file does not exist
		 * @return 0 on success or an error code
		 */
		int importFromMetaFile( const std::string& study, const std::string& file, bool& missingMeta );

		/**
		 * Performs the actual import by reading the objectdb.meta file and inserting to the Dcobject table.
		 * Looks up the objectdb.meta file and calls @link #importFromMetaFile.
		 *
		 * @param study the study instance UID
		 * @param missingMeta output parameter to indicate that the object meta file does not exist
		 * @return 0 on success or an error code
		 */
		int importFromMeta( const std::string& study, bool& missingMeta );

	private:

		/**
		 * Checks whether the object table entries of a particular study are mapped to the database.
		 * Maps the object table entries if needed.
		 *
		 * Also touches the study in the object cache table regardless to whether the object entries were already mapped or not.
		 *
		 * When a new order gets created or a study is just received, the object meta file may not yet exist, but we still need to create
		 * an object cache entry.
		 * However a missing object meta file is also a valid  when a study is being deleted,
		 * in which case we do not want to recreate the object cache entry.
		 * Whether an object cache entry to be created when there is no object meta file is controlled by the <code>mapMissing</code> parameter.
		 *
		 * @param study the study instance UID
		 * @param mapMissing if <code>true</code>, then an object cache entry will be created even if no object meta file exists.
		 * @throw IMTimeoutException if could not get lock to import the object table entries
		 * @throw ObjCacheException if an error occurs
		 */
		void doTouch( const std::string& study, bool mapMissing ) noexcept ( false );

		/**
		 * Collects the object file names from the objectdb.meta file (if any)
		 *
		 * @param study the study instance UID
		 * @param files output parameter for the collected file names
		 * @return 0 on success or an error code
		 */
		int getFilesFromMeta( const std::string& study, std::set<std::string>& files );

		/**
		 * Adds 'ocerror' references to all objects in the study.
		 * Collects object file names from the object table, the dicom repository and the objectdb.meta file.
		 *
		 * @param study the study instance UID
		 */
		void addErrorRefs( const std::string& study );

		/**
		 * Collects the object file names from the repository locations of the resource.
		 * Tries to find the resource locations on all available mount points (even if there is a repository conflict)
		 *
		 * @param resource the resource to find
		 * @param locations output parameter for the locations found
		 * @param objects output parameter for the object file names collected from the locations
		 */
		void getRepoObjects( const std::string& resource, std::list<std::string>& locations, std::set<std::string>& objects );


		/// Database connection
		erad::db::DBConnectorSql* m_pDB;
		/// The database name of the object table
		std::string objectTbl;
		/// The database name of the object cache table
		std::string objectCacheTbl;
		/// Path to the DICOM repository
		std::string dicomRepository;
		/// logger for error/debug logging
		IMLogger logger;
};

} // namespace

#endif //OBJCACHEMGR_H
