/*
 * Component: utils
 *
 * Description:
 *         A simpler interface for adding/removing references from C/C++ programs.
 *
 */

#ifndef IM_REF_COUNT_H_INCLUDED
#define IM_REF_COUNT_H_INCLUDED

#include "utils/ReferenceCounter.h"
#include "imutils/DBConnector.h"


/**
 * A simplier wrapper interface for adding/removing references from C/C++ programs.
 *
 * addReference() and removeReference() methods will load the reference counter file, add/remove the reference
 * and then release the reference counter file.
 *
 * Both methods will catch exceptions, and returns error (if applicable) in such cases.
 *
 * It's worth to keep a static RefCount object in the memory and use it to manipulate reference counters.
 */
class RefCount {

		erad::db::DBConnector& db;

		// Release configuration file path for the reference counters
		OFString m_strRelConfigPath;

		// Verbose mode for printing verbose messages
		bool m_verbose;

	public:

		/**
		 * Constructor for creating a new RefCount object. It will load the release configuration file.
		 *
		 * @param verbose   verbose messages will be printed on standard output if true (false by default)
		 */
		RefCount( erad::db::DBConnector& db, bool verbose = false );

		/**
		 * Adds a reference to a reference counter
		 *
		 * @param referenceType   type of the reference counter
		 * @param refcountId      identifies the reference counter
		 * @param reference       reference to be added
		 * @param bCreateIfNew	  if set to false, call will fail if refount object does not exist
		 * 							  instead of creating a new one (true by default!)
		 *
		 * @return number of references added (1 or 0) on success,
		 * 			-2 when adding to non-existing refcounter with bCreateIfNew set to false,
		 * 			-3 on fs or -4 on db failure, -5 on underlying IMExcpetion
		 */
		int addRef( const char* referenceType, const char* refcountId, const char* reference, bool bCreateIfNew = true );

		/**
		 * Simpler method to add reference to a reference counter,
		 * creates new refcount object if it did not exists.
		 *
		 * @param referenceType   type of the reference counter
		 * @param refcountId      identifies the reference counter
		 * @param reference       reference to be added
		 *
		 * @return true if and only if the reference has been successfully added
		 */
		inline bool addReference( const char* referenceType, const char* refcountId, const char* reference ) {
			return ( 0 < addRef( referenceType, refcountId, reference, true ) );
		}

		/**
		 * Adds a new reference to a reference counter ONLY if the reference does not yet exist
		 *
		 * @param referenceType   type of the reference counter
		 * @param refcountId      identifies the reference counter
		 * @param reference       reference to be added
		 * @param bCreateIfNew	  if set to false, call will fail if refount object does not exist
		 * 							  instead of creating a new one (true by default!)
		 *
		 * @return number of references added (1 or 0) on success,
		 * 			-2 when adding to non-existing refcounter with bCreateIfNew set to false,
		 * 			-3 on fs or -4 on db failure, -5 on underlying IMExcpetion
		 */
		int newRef( const char* referenceType, const char* refcountId, const char* reference, bool bCreateIfNew = true );

		/**
		 * Removes a reference from a reference counter
		 *
		 * @param referenceType   type of the reference counter
		 * @param refcountId      identifies the reference counter
		 * @param reference       reference to be added
		 *
		 * @return number of references removed (1 or 0) on success,
		 * 			-2 if refcounter does not exist
		 * 			-3 on fs or -4 on db failure, -5 on underlying IMExcpetion
		 */
		int remRef( const char* referenceType, const char* refcountId, const char* reference );

		inline void removeReference( const char* referenceType, const char* refcountId, const char* reference ) {
			remRef( referenceType, refcountId, reference );
		}
};

#endif // IM_REF_COUNT_H_INCLUDED
