/*
 * Component:  utils
 *
 * Purpose:
 *   Header file for reference counter implementation
 */

#ifndef REFLOCKENTRY_H
#define REFLOCKENTRY_H

#include <stdio.h>

#include <dcmtk/ofstd/oflist.h>
#include <dcmtk/ofstd/ofstring.h>


class ReleaseConfigEntry;


/**
 * Reads configuration file of registered commands to be started when an object is released.
 */
class ReleaseConfig {
	private:

		/** List of configuration file entries */
		OFList<ReleaseConfigEntry*> m_entryList;

	public:

		/**
		 * Constructor for creating a new ReleaseConfig object
		 *
		 * @param configFile    name of the configuration file (if empty tries REFERENCE_RELEASE_CONFIG)
		 */
		ReleaseConfig();

		/**
		 * Destructor for destroying a ReleaseConfig object
		 */
		~ReleaseConfig();

		/**
		 * Reads the configuration file
		 */
		void readConfigFile( const OFString& configFile );

		/**
		 * Checks if there is any command is registered for this type and starts
		 * the registered commands. id will be passed to the command
		 *
		 * @param type   check commands registered for this object type
		 * @param id     ID of the object. Will be passed to the commands.
		 */
		void checkAndStartCommandsFor( const OFString& type, const OFString& event, const OFString& id );
};


class ReferenceCounterImplBase;


/**
 * Reference counter implementation. A reference counter may contain several
 * referrer entries for an object. Each referrer may refer the object several times.
 * Reference counters are stored in a file. If there is no reference on the object the
 * reference counter file will be removed.
 *
 * @note For simple manipulation of reference counters from C/C++ programs you might want to use RefCount objects.
 */
class ReferenceCounter {
		ReferenceCounterImplBase* m_pImpl; // hehe: PIMP(l), no actually PIMPL pattern
	public:
		/**
		 * Constructor for creating a reference counter object
		 *
		 * @param id          identifies the object to be referenced
		 * @param type        type of the object to be referenced (eg. STUDY, OBJECT)
		 */
		ReferenceCounter( const OFString& id, const OFString& type );

		ReferenceCounter( void* pDB, const OFString& id, const OFString& type );

		/**
		 * Destructor for destroying a reference counter object
		 */
		~ReferenceCounter();

	public:
		/**
		 * Retrieves the list of refernces and the comment (atomic!)
		 *
		 * @param referrers 	(out) list to populate with referrer ids;
		 * @param comment 	(out) string to contain comment;
		 *
		 * @return false if refcount does not exis, true otherwise
		 */
		bool getInfo( OFList<OFString>& referrers /*out*/, OFString& comment /*out*/);

		/**
		 * Serializes refcounter information to stream (text),
		 * used internally by the filesystem based engine (write), but
		 * can be called externally to format arbitrary parameters
		 *
		 * @param file   			file stream to output format data to
		 * @param type   			type of the object
		 * @param comment 		comment
		 * @param references  	list of referrers
		 */
		static void stream( FILE* file, const OFString& type, const OFString& comment, const OFList<OFString>& references );

		/**
		 * Cleans up the object held by the reference counter when the last reference get removed from the reference counter
		 * (the reference counter get released)
		 */
		void onRelease();

	public:
		/**
		 * batch modify references (atomic!)
		 *
		 * @param mode  		access mode: (default: 0; create only: 1; existing only: 2; any: 3)
		 * @param comment 	pointer to comment string, field will not be updated if NULL
		 * @param toAdd 		list of references to add
		 * @param toCreate	list of references to create (only if they don't exist)
		 * @param toRemove	list of references to remove
		 *
		 * @return (positive or zero) the number of entries changed (including comment change),
		 * 			or
		 * 			(negitive) error code:
		 * 			-1: access mode 1 violated
		 * 			-2: access mode 2 violated
		 * 			-3: file save error
		 * 			-4: database error
		 *
		 */
		int batchProcess( const int mode, const OFString* comment,
				const OFList<OFString>& toAdd,
				const OFList<OFString>& toCreate,
				const OFList<OFString>& toRemove );
};


/**
 * Entries in the configuration file of registered commands to be started when an object is released.
 */
class ReleaseConfigEntry {
	private:
		/** Object type */
		OFString m_type;

		/** Event (RELEASED) */
		OFString m_event;

		/** Command to be executed */
		OFString m_command;

	public:
		/**
		 * Creates a new configuration file entry
		 *
		 * @param type     Object type
		 * @param event    Event (RELEASED)
		 * @param command  Command to be executed
		 */
		ReleaseConfigEntry( const OFString& type, const OFString& event, const OFString& command );

		/**
		 * Destructor for destroying a ReleaseConfig object
		 */
		~ReleaseConfigEntry();

		/**
		 * Checks if this configuration entry is matching to a type
		 *
		 * @param type  check configuration entry against this type
		 *
		 * @return true if and only if this configuration entry is matching to the type
		 */
		inline bool isMatchingType( const OFString& type ) {
			return m_type.compare( type ) == 0;
		}

		/**
		 * Checks if this configuration entry is matching to an event
		 *
		 * @param event    check configuration entry against this event
		 *
		 * @return true if and only if this configuration entry is matching to the event
		 */
		inline bool isMatchingEvent( const OFString& event ) {
			return m_event.compare( event ) == 0;
		}

		/**
		 * Checks if this configuration entry is matching to a type and an event
		 *
		 * @param type   check configuration entry against this type
		 * @param event  check configuration entry against this event
		 *
		 * @return true if and only if this configuration entry is matching to both the type and the event
		 */
		inline bool isMatching( const OFString& type, const OFString& event ) {
			return isMatchingType( type ) && isMatchingEvent( event );
		}

		/**
		 * Retrieves the command to be executed.
		 *
		 * @return  the registered command to be executed
		 */
		inline const OFString& getCommand() {
			return m_command;
		}
};

#endif
