/*
 * Component:  utils
 *
 * Purpose: 
 *   Header file for manipulating storestate.rec files
 *
 */
#ifndef __STORESTATE_H_INCLUDED_
#define __STORESTATE_H_INCLUDED_

#include <stdio.h>

// #include <dcmtk/config/osconfig.h>
#include <dcmtk/ofstd/ofstring.h>
#include <dcmtk/ofstd/ofuoset.h>
#include <dcmtk/ofstd/oflist.h>

#include "utils/MetaMgr.h"

#include "imutils/IMSemaphore.h"
#include "imutils/DBConnector.h"


#define CURRENT_STORESTATE_VERSION 1

/**
 * Class for reading and manipulating storestate.rec files of studies.
 *
 * storestate.rec file attributes:
 *
 * OwnerDevice:       Next dotcom device toward the study owner
 * ProcessMode:       How to handle incoming image objects - Process/Store/Deleted/Database/ReportOnly
 * StagedState:       'Staged' indicates if study retrieve was initiated 
 * Folder:            Name of the folder the study belongs to (for studies copied into a folder)
 * SourceDevice:      Identifies the device the study was received from (filled for archives)
 * ArchiveDevice:     Id of the dotcom device where the study to be archived
 * StagedDevices:     List of neighbour dotcom devices where the study to be forwarded
 * ReportOnlyDevices: List of neighbour dotcom devices where the report files (only) to be forwarded
 * ForwardDevices:    List of non dotcom devices where the study to be forwarded
 * SourceStudy:       Study Instance UID of the original study a link study refers to
 * StudyOwner:        'yes' indicates that the current device is the study owne/
 * Retrieved:         'yes' indicates that the study is retrieved from an archive (either internal or external)
 * DontEdit:          'yes' indicates that the study is not editable
 * DontDelete:        'yes' indicates that the study is not deletable (?? used for orders ?? obsolete ??)
 * Derived:           Filled for derived studies, the value is the operation created the study (eg.: 'merge', 'split', 'copy')
 * Version:           Version of the storestate.rec file
 *
 * Exactly one of StudyOwner (='yes') and OwnerDevice (not empty) attributes should exist.
 * A dotcom device can be either OwnerDevice, StagedDevice or ReportOnlyDevice but not both.
 */
class StoreState
{
private:
	
	/**
	 * Inner class used to store AutoForwardDevices 
	 */
	class StoreDevice{
		private:
			OFString m_device;
			OFString m_type;

		public:
			/**
			 *Constructor for StoreDevice class.
			 *
			 * @param device the name of the device
			 * @param type the type of the device
			*/			
			StoreDevice(const OFString & device, const OFString & type);
			
			/**
			 * Get the name of the device 
			 *
			 * @return get the name of the device
			 */
			const OFString & getDevice() {
				return m_device;
			}
			
			/**
			 * Set the name of the device
			 *
			 * @param p_device the name of the device 
			 */
			void setDevice(const OFString & p_device) {
				m_device = p_device;
			}
			
			
			/**
			 * Get the type of the device
			 *
			 * @return get the type of the device
			 */  
			const OFString & getType() {
				return m_type;
			}
			
			/**
			 * Set the type of the device
			 *
			 * @param p_device the type of the device
			 */ 
			void setType(const OFString & p_type) {
				m_type = p_type;
			}
	};
	

	/** OwnerDevice: Next device toward the study owner */
	OFString m_ownerDevice;

	/** ProcessMode: How to handle incoming image objects - Process/Store/Deleted/Database/ReportOnly */ 
	OFString m_processMode;

	/** StagedState: 'Staged' indicates if study retrieve was initiated */
	OFString m_stageState;

	/** Folder: Name of the folder the study belongs to (for studies copied into a folder) */
	OFString m_folder;

	/** SourceDevice: Identifies the device the study was received from (filled for archives) */
	OFString m_sourceDevice;

	/** ArchiveDevice: Id of the dotcom device where the study to be archived */
	OFString m_archiveDevice;

	/** StagedDevices: List of neighbour dotcom devices where the study to be forwarded */
	OFUnorderedSet<OFString> m_stagedDevices;

	/** ReportOnlyDevices: List of neighbour dotcom devices where the report files (only) to be forwarded */
	OFUnorderedSet<OFString> m_reportOnlyDevices;

	/** ForwardDevices: List of non dotcom devices where the study to be forwarded */
	OFUnorderedSet<OFString> m_forwardDevices;

	/** Name of the operation used for creating the study from another studies (eg: 'merge', 'split', 'copy') */
	OFString m_derived;

	/** StudyOwner: 'yes' indicates that the current device is the study owner */
	bool m_isStudyOwner;

	/** Retrieved: 'yes' indicates that the study is retrieved from an archive (either internal or external) */
	bool m_isRetrieved;

	/** DontEdit: 'yes' indicates that the study is not editable */
	bool m_isDontEdit;

	/** DontDelete: 'yes' indicates that the study is not deletable (?? used for orders ?? obsolete ??) */
	bool m_isDontDelete;

	/** Version: version of the storestate.rec file */
	int m_version;

	/** Study Instance UID of the study the storestate.rec file belongs to */	
	OFString m_study;

	/** Study Instance UID of the original study a link study refers to */
	OFString m_sourceStudy;
	
	/** managing meta directories */
	erad::utils::MetaManager metaManager;

	/** Semaphore for getting exclusive lock when manipulating storestate.rec files */
	IMSemaphore m_semaphore;

	/** Variable for storage of AutoForwardDevices.**/
	OFList<StoreState::StoreDevice*> m_autoFDList;

	/** Database connection */
	erad::db::DBConnector *dbConn;

	/** Controls whether to discard the database connection on cleanup */
	bool discardDbConn;

	/** */
	bool m_needLock;

	/**
	* Parse buffer for a string value. Mask should contain exactly one char * reference.
	*
	* @param buffer    string buffer to read from
	* @param mask      mask for sscanf
	* @param var       output parameter to store parsed string value
	*
	* @return true if and only if property value was successfully parsed
	*/
	bool readProperty(const char * buffer, const char * mask, char * var);

	/**
	 * Parses space separated list of values (device IDs) and puts each value to a set.
	 *
	 * @param set  to store individual values in
	 * @param str  space separated values to be parsed
	 */
	void createSetFromOFString(OFUnorderedSet<OFString> & set, const OFString & str);

	/**
	 * Adds devices flagged to be AutoForward in the database to ForwardDevices attribute.
	 * If called from convert(), study owner flag will be checked and set if neither StudyOwner is set 
	 * nor Parent device is configured.
	 *
	 * @param isConvert  true means called from convert() (converting old storestate.rec files)
	 *
	 * @throws IMException if an error occurs
	 */
	void addAutoForwardDevices(bool isConvert);

	/**
	 * Converts (updates) an old storestate.rec file to the most recent format (version).
	 *
	 * @throws IMException if an error occurs
	 */
	void convert();

public:

	/**
	 * Set AutoForwardDevices from command line.
	 * 
	 * @param cmdArgument the argument from the command line.
	 * 		 The format: DeviceName.DeviceType;DeviceName.DeviceType;
	 * 		 The ; in the end of the argument is necessary.
	 * */
	void setAutoForwardDevice(const char* cmdArgument);
	
	StoreState( const OFString &study );

	StoreState( erad::db::DBConnector *db, const OFString &study );

	/**
	 * Destructor for destroying Storestate object.
	 */
	~StoreState();


	/**
	 * Sets custom lock directory (for testing).
	 * Should be set immediately after construction.
	 *
	 * @param lockDir path to the lock directory
	 */
	inline void setLockDir( const OFString& lockDir ) {
		m_semaphore.setLockDir( lockDir );
	}


	/**
	 * Tests whether the storestate.rec file exists.
	 * @return true if and only if the storestate.rec file exists; false otherwise
	 * @throw IMException  if an error occurs
	 */
	bool exists();

	/** 
	 * Loads the storestate.rec file. Also locks the storestate.rec file
	 * for modifications except the locking is turned off using setNeedLock().
	 * 
	 * @return true   if the storestate.rec file exists and has been successfully read
	 *         false  if the storestate.rec file does not exist
	 *
	 * @throws IMIOException       if I/O error occurs
	 *         IMTimeoutException  if can't get lock for storestate.rec file within timeout
	 *         IMException         if an error occurs
	 */
	bool load();

	/**
	 * Unlocks the storestate.rec file without saving the changes.
	 *
	 * @throws IMIOException if an I/O error occurs
	 */
	void release();

	/**
	 * Saves changes and unlocks the storestate.rec file.
	 *
	 * @return true if and only if changes were successfully saved
	 * 
	 * @throws IMIOException       if I/O error occurs
	 *         IMException         if an error occurs
	 */
	bool save();    // save and release

	/**
	 * Removes (deletes) the storestate.rec file.
	 *
	 * @return true if the storestate.rec file has been successfully removed
	 *
	 * @throws IMException if an error occurs
	 */
	bool remove();

	/**
	 * Prints the content of the storestate.rec file to file stream.
	 *
	 * @param f   file stream to print the content of the storestate.rec file to.
	 */
	void print(FILE * f);

	/**
	 * Sets the needLock flag.
	 * If the needLock flag is 'true' (default), the load() will lock the storestate.rec
	 * file and save() will releas it.
	 * If the needLock flag is 'false', the storestate.rec file is supposed to be locked using
	 * external mechanism, and manipulation is enabled without locking.
	 *
	 * @param needLock   new value for the needLock flag
	 */
	inline void setNeedLock(bool needLock) {
		m_needLock = needLock;
	}

	/**
	 * Returns the OwnerDevice attribute. Empty string means no OwnerDevice is set.
	 *
	 * @return OwnerDevice attribute
	 */
	inline const OFString & getOwnerDevice() const {
		return m_ownerDevice;
	}

	/**
	 * Sets the OwnerDevice attribute.
	 * Also sets the StudyOwner flag accordingly.
	 *
	 * @param device  the new owner device
	 */
	inline void setOwnerDevice(const OFString & device) {
		m_ownerDevice = device;
		m_isStudyOwner = m_ownerDevice.empty();
	}

	/**
	 * Clears the OwnerDevice attribute.
	 * Also sets the StudyOwner flag to 'yes'.
	 */
	inline void clearOwnerDevice() {
		m_ownerDevice = ""; m_isStudyOwner = true;
	}

	/**
	 * Returns the ProcessMode attribute.
	 *
	 * @return ProcessMode attribute
	 */
	inline const OFString & getProcessMode() const {
		return m_processMode;
	}

	/**
	 * Sets the ProcessMode attribute
	 *
	 * @param processMode  the new value of process mode
	 */
	inline void setProcessMode(const OFString & processMode) {
		m_processMode = processMode;
	}

	/**
	 * Returns the StageState attribute.
	 *
	 * @return StageState attribute
	 */
	inline const OFString & getStageState() const {
		return m_stageState;
	}

	/**
	 * Sets the StageState attribute
	 *
	 * @param StageState the new value of process mode
	 */
	inline void setStageState(const OFString & stageState) {
		m_stageState = stageState;
	}

	/**
	 * Returns the Folder attribute.
	 *
	 * @return Folder attribute
	 */
	inline const OFString & getFolder() const {
		return m_folder;
	}

	/**
	 * Sets the Folder attribute
	 *
	 * @param Folder the new value of process mode
	 */
	inline void setFolder(const OFString & folder) {
		m_folder = folder;
	}

	/**
	 * Returns the SourceDevice attribute.
	 *
	 * @return SourceDevice attribute
	 */
	inline const OFString & getSourceDevice() const {
		return m_sourceDevice;
	}

	/**
	 * Sets the SourceDevice attribute
	 *
	 * @param SourceDevice the new value of process mode
	 */
	inline void setSourceDevice(const OFString & sourceDevice) {
		m_sourceDevice = sourceDevice;
	}

	/**
	 * Returns the ArchiveDevice attribute.
	 *
	 * @return ArchiveDevice attribute
	 */
	inline const OFString & getArchiveDevice() const {
		return m_archiveDevice;
	}

	/**
	 * Sets the ArchiveDevice attribute
	 *
	 * @param ArchiveDevice the new value of process mode
	 */
	inline void setArchiveDevice(const OFString & archiveDevice) {
		m_archiveDevice = archiveDevice;
	}

	/**
	 * Add a dotcom device to list of staged devices.
	 *
	 * @param device  Id of the dotcom device to be added to the list of staged devices
	 */
	inline void addStagedDevice(const OFString & device) {
		if (!m_stagedDevices.Contains(device))
			m_stagedDevices.Insert(device);
	}

	/**
	 * Remove a dotcom device from the list of staged devices.
	 * 
	 * @param device    Id of the dotcom device to be removed from the list of staged devices
	 */	
	inline void removeStagedDevice(const OFString & device) {
		m_stagedDevices.Remove(device);
	}


	/**
	 * Returns a read only reference to staged devices
	 *
	 * @return	The const reference to the OFUnorderedset 
	 */
	inline const OFUnorderedSet<OFString> & getStagedDevices() {
		return m_stagedDevices;
	}	

	inline const OFUnorderedSet<OFString> & getForwardDevices() {
		return m_forwardDevices;
	}

	
	/**
	 * Checks if a dotcom device is a staged device (list of staged devices contains the dotcom device).
	 *
	 * @param device    Id of the dotcom device to be checked.
	 * 
	 * @return true if and only if the dotcom device is a staged device.
	 */
	inline bool isStagedDevice(const OFString & device) const {
		return m_stagedDevices.Contains(device);
	}

	/**
	 * Add a dotcom device to list of report only devices.
	 *
	 * @param device  Id of the dotcom device to be added to the list of report only devices
	 */
	inline void addReportOnlyDevice(const OFString & device) {
		if (!m_reportOnlyDevices.Contains(device))
			m_reportOnlyDevices.Insert(device);
	}

	/**
	 * Remove a dotcom device from the list of report only devices.
	 * 
	 * @param device    Id of the dotcom device to be removed from the list of report only devices
	 */	
	inline void removeReportOnlyDevice(const OFString & device) {
		m_reportOnlyDevices.Remove(device);
	}

	/**
	 * Checks if a dotcom device is a report only device (list of report only devices contains the dotcom device).
	 *
	 * @param device    Id of the dotcom device to be checked.
	 * 
	 * @return true if and only if the dotcom device is a report only device.
	 */
	inline bool isReportOnlyDevice(const OFString & device) const {
		return m_reportOnlyDevices.Contains(device);
	}

	/**
	 * Add a device to list of forward devices.
	 *
	 * @param device  Id of the device to be added to the list of forward devices
	 */
	inline void addForwardDevice(const OFString & device) {
		if (!m_forwardDevices.Contains(device))
			m_forwardDevices.Insert(device);
	}

	/**
	 * Remove a device from the list of forward devices.
	 * 
	 * @param device    Id of the device to be removed from the list of forward devices
	 */	
	inline void removeForwardDevice(const OFString & device) {
		m_forwardDevices.Remove(device);
	}

	/**
	 * Checks if a device is a forward device (list of forward devices contains the device).
	 *
	 * @param device    Id of the device to be checked.
	 * 
	 * @return true if and only if the device is a forward device.
	 */
	inline bool isForwardDevice(const OFString & device) const {
		return m_forwardDevices.Contains(device);
	}

	/**
	 * Sets the name of the operation the study was created by for a derived study
	 *
	 * @param derived   the name of the operation the study was created by
	 */
	inline void setDerived( const OFString & derived ) {
		m_derived = derived;
	}

	/**
	 * Returns the name of the operation the study was created by for derived studies.
	 * Returns empty string for non derived studies (ie. studies was received from modality or retreived from archive)
	 *
	 * @return    the name of the operation the study was created by
	 */
	inline const OFString & getDerived() const {
		return m_derived;
	}

	
	/**
	 * Checks if the current device is the study owner (StudyOwner attribute is 'yes').
	 *
	 * @return true if and only if the current device is the study owner
	 */
	inline bool isStudyOwner() const {
		return m_isStudyOwner;
	}

	/**
	 * Sets the StudyOwner attribute.
	 * If the new value is 'yes', also clears the OwnerDevice attribute.
	 *
	 * @param isStudyOwner   true indicates to set the StudyOwner attribute to 'yes'
	 */
	inline void setStudyOwnerFlag(bool isStudyOwner) {
		m_isStudyOwner = isStudyOwner;
		if (m_isStudyOwner)
			m_ownerDevice = "";
	}

	/**
	 * Checks if the study is a retrieved study (Retrieved attribute is 'yes').
	 *
	 * @return true if and only if the study is a retrived study.
	 */
	inline bool isRetrieved() const {
		return m_isRetrieved;
	}

	/**
	 * Sets the Retrieved attribute.
	 *
	 * @param isRetrieved  true means to set the Retrieved attribute to 'yes'
	 */
	inline void setRetrievedFlag(bool isRetrieved) {
		m_isRetrieved = isRetrieved;
	}

	/**
	 * Checks if the DontEdit attribute is set for the study.
	 *
	 * @return true if and only if the DontEdit attribute is set for the study.
	 */
	inline bool isDontEdit() const {
		return m_isDontEdit;
	}

	/**
	 * Sets the DontEdit attribute.
	 *
	 * @param isDontEdit  true means to set the DontEdit attribute to 'yes'
	 */
	inline void setDontEditFlag(bool isDontEdit) {
		m_isDontEdit = isDontEdit;
	}

	/**
	 * Checks if the DontDelete attribute is set for the study.
	 *
	 * @return true if and only if the DontDelete attribute is set for the study.
	 */
	inline bool isDontDelete() const { return m_isDontDelete; }
	
	/**
	 * Sets the DontDelete attribute.
	 *
	 * @param isDontDelete  true means to set the DontDelete attribute to 'yes'
	 */
	inline void setDontDeleteFlag(bool isDontDelete) {
		m_isDontDelete = isDontDelete;
	}

	/**
	 * Returns the SourceStudy attribute (original Study for a link study).
	 *
	 * @return Original Study Instance UID for a link study
	 */
	inline const OFString & getSourceStudy() const {
		return m_sourceStudy;
	}

	/**
	 * Sets the SourceStudy attribute (original Study for a link study).
	 * 
	 * @param sourceStduy  Original Study Instance UID for a link study
	 */
	inline void setSourceStudy(const OFString & sourceStudy) {
		m_sourceStudy = sourceStudy;
	}
	
	/**
	 * Gets the semaphore used for locking the storestate.rec file.
	 *
	 * @return the semaphore used for locking the storestate.rec file
	 */
	inline IMSemaphore & getSemaphore() {
		return m_semaphore;
	}
};


#endif // __STORESTATE_H_INCLUDED_
