/*
 *  Module:  cd
 *
 *  Author:  Tamas Nemeth
 *
 *  Purpose: Interface of RLECodec and the RLECodecParameter
 *
 *  Last Update:      $Author$
 *  Update Date:      $Date$
 *  Source File:      $Source$
 *  CVS/RCS Revision: $Revision$
 *  Status:           $State$
 *
 *  CVS/RCS Log at end of file
 *
 */

#ifndef RLECODEC_H
#define RLECODEC_H

// #include <dcmtk/config/osconfig.h>    /* make sure OS specific configuration is included first */
#include <dcmtk/dcmdata/dccodec.h>

class RLECodecParameter : public DcmCodecParameter
{
public:
    RLECodecParameter();

    RLECodecParameter(const RLECodecParameter&);

    /// destructor
    virtual ~RLECodecParameter();
    
    /** this methods creates a copy of type RLECodecParameter *
     *  it must be overweritten in every subclass.
     *  @return copy of this object
     */
    virtual RLECodecParameter *clone() const;

    /** returns the class name as string.
     *  can be used as poor man's RTTI replacement.
     */
    virtual const char *className() const;
};                    


/** abstract base class for a codec object that can be registered
 *  in dcmdata and performs transfer syntax transformation (i.e.
 *  compressing, decompressing or transcoding between different 
 *  compressed transfer syntaxes).
 *  When dcmdata is requested to write a transfer syntax that differs
 *  from the current one (i.e. the one in which the object was read),
 *  dcmdata dynamically searches for a DcmCodec object that is able
 *  to create the desired transfer syntax.  If no suitable codec
 *  is found, the write operation fails.
 */
class RLECodec : public DcmCodec
{
public:
  /// default constructor
  RLECodec();
  virtual ~RLECodec();

  /** decompresses the given pixel sequence and
   *  stores the result in the given uncompressedPixelData element.
   *  @param fromRepParam current representation parameter of compressed data, may be NULL
   *  @param pixSeq compressed pixel sequence
   *  @param uncompressedPixelData uncompressed pixel data stored in this element
   *  @param cp codec parameters for this codec
   *  @param objStack stack pointing to the location of the pixel data
   *    element in the current dataset.
   *  @return EC_Normal if successful, an error code otherwise.
   */
  virtual OFCondition decode(
      const DcmRepresentationParameter * fromRepParam,
      DcmPixelSequence * pixSeq,
      DcmPolymorphOBOW& uncompressedPixelData,
      const DcmCodecParameter * cp,
      const DcmStack& objStack) const;

   /** decompresses a single frame from the given pixel sequence and
    *  stores the result in the given buffer.
    *  @param fromParam representation parameter of current compressed
    *    representation, may be NULL.
    *  @param fromPixSeq compressed pixel sequence
    *  @param cp codec parameters for this codec
    *  @param dataset pointer to dataset in which pixel data element is contained
    *  @param frameNo number of frame, starting with 0 for the first frame
    *  @param startFragment index of the compressed fragment that contains
    *    all or the first part of the compressed bitstream for the given frameNo.
    *    Upon successful return this parameter is updated to contain the index
    *    of the first compressed fragment of the next frame.
    *    When unknown, zero should be passed. In this case the decompression
    *    algorithm will try to determine the index by itself, which will always
    *    work if frames are decompressed in increasing order from first to last,
    *    but may fail if frames are decompressed in random order, multiple fragments
    *    per frame and multiple frames are present in the dataset, and the offset
    *    table is empty.
    *  @param buffer pointer to buffer where frame is to be stored
    *  @param bufSize size of buffer in bytes
    *  @param decompressedColorModel upon successful return, the color model
    *    of the decompressed image (which may be different from the one used
    *    in the compressed images) is returned in this parameter.
    *  @return EC_Normal if successful, an error code otherwise.
    * 
    * THIS IS A DUMMY IMPLEMENTATION ALWAYS RETURNING 
    */
  virtual OFCondition decodeFrame(
    const DcmRepresentationParameter * fromParam,
    DcmPixelSequence * fromPixSeq,
    const DcmCodecParameter * cp,
    DcmItem *dataset,
    Uint32 frameNo,
    Uint32& startFragment,
    void *buffer,
    Uint32 bufSize,
    OFString& decompressedColorModel) const;
    
      
  /** compresses the given uncompressed DICOM image and stores
   *  the result in the given pixSeq element.
   *  @param pixelData pointer to the uncompressed image data in OW format
   *    and local byte order
   *  @param length of the pixel data field in bytes
   *  @param toRepParam representation parameter describing the desired
   *    compressed representation (e.g. JPEG quality)
   *  @param pixSeq compressed pixel sequence (pointer to new DcmPixelSequence object
   *    allocated on heap) returned in this parameter upon success.   
   *  @param cp codec parameters for this codec
   *  @param objStack stack pointing to the location of the pixel data
   *    element in the current dataset.
   *  @return EC_Normal if successful, an error code otherwise.
   */
  virtual OFCondition encode(
      const Uint16 * pixelData,
      const Uint32 length,
      const DcmRepresentationParameter * toRepParam,
      DcmPixelSequence * & pixSeq,
      const DcmCodecParameter *cp,
      DcmStack & objStack) const;

  /** transcodes (re-compresses) the given compressed DICOM image and stores
   *  the result in the given toPixSeq element.
   *  @param fromRepType current transfer syntax of the compressed image
   *  @param fromRepParam current representation parameter of compressed data, may be NULL
   *  @param fromPixSeq compressed pixel sequence
   *  @param toRepParam representation parameter describing the desired
   *    new compressed representation (e.g. JPEG quality)
   *  @param toPixSeq compressed pixel sequence (pointer to new DcmPixelSequence object
   *    allocated on heap) returned in this parameter upon success.   
   *  @param cp codec parameters for this codec
   *  @param objStack stack pointing to the location of the pixel data
   *    element in the current dataset.
   *  @return EC_Normal if successful, an error code otherwise.
   */
  virtual OFCondition encode(
      const E_TransferSyntax fromRepType,
      const DcmRepresentationParameter * fromRepParam,
      DcmPixelSequence * fromPixSeq,
      const DcmRepresentationParameter * toRepParam,
      DcmPixelSequence * & toPixSeq,
      const DcmCodecParameter * cp,
      DcmStack & objStack) const;

  /** checks if this codec is able to convert from the
   *  given current transfer syntax to the given new
   *  transfer syntax
   *  @param oldRepType current transfer syntax
   *  @param newRepType desired new transfer syntax
   *  @return true if transformation is supported by this codec, false otherwise.
   */
  virtual OFBool canChangeCoding(
      const E_TransferSyntax oldRepType,
      const E_TransferSyntax newRepType) const;

  /** determine color model of the decompressed image
   *  @param fromParam representation parameter of current compressed
   *    representation, may be NULL
   *  @param fromPixSeq compressed pixel sequence
   *  @param cp codec parameters for this codec
   *  @param dataset pointer to dataset in which pixel data element is contained
   *  @param dataset pointer to DICOM dataset in which this pixel data object
   *    is located. Used to access photometric interpretation.
   *  @param decompressedColorModel upon successful return, the color model
   *    of the decompressed image (which may be different from the one used
   *    in the compressed images) is returned in this parameter
   *  @return EC_Normal if successful, an error code otherwise
   */
  virtual OFCondition determineDecompressedColorModel(
    const DcmRepresentationParameter *fromParam,
    DcmPixelSequence *fromPixSeq,
    const DcmCodecParameter *cp,
    DcmItem *dataset,
    OFString &decompressedColorModel) const;

};

#endif

/*
** CVS/RCS Log:
** $Log$
** Revision 1.2  2006/09/27 09:07:28  tnemeth
** Upgrade to DCMTK-3.5.4
**
** Revision 1.1  2002/10/08 11:35:39  tnemeth
** Initial revision
**
** Revision 1.1  2002/09/13 13:53:58  tnemeth
** Initial revision
**
*/
