/**
 * Component: utils
 *
 * Description:
 *    Helper methods to start tasks
 *
 *  TODO: on long term these methods should be moved to component taskd
 *
 */

#ifndef TASKHELPER_H_
#define TASKHELPER_H_

#include "dcmtk/ofstd/ofstd.h"
#include "imutils/DBConnector.h"


/**
 * Reindex modes
 */
typedef enum {
	/** Regular store - no reindex */
	STORESCP_REG_OPT_NONE,
	/** Reindex to the database only */
	STORESCP_REG_OPT_REINDEX,
	/** Reindex and autoforward, ...*/
	STORESCP_REG_OPT_FULL,
	/** Called from dcfixall (implies full)*/
	STORESCP_REG_OPT_DCFIXALL
} STORESCP_REG_OPT;


/**
 * Provides enumerated values controlling how a DICOM object to be registered
 * @enum RegistrationMode
 */
enum class RegistrationMode {
	/** Register the object file directly, fail, if registration fails */
	direct,
	/** Register the object by scheduling a task */
	task,
	/** Try to register the object file directly, scheduling a task if registering directly fails */
	directOrTask
};


bool runJavaTask( const std::string& cmd );


class StorescpRegStarter {
	private:

		/** Instance UID of the study the object to be stored for */
		OFString m_study;
		/** File name of the object */
		OFString m_file;
		/** Reindex mode */
		STORESCP_REG_OPT m_reindexMode;
		/** Path to the object file to be stored */
		OFString m_tempFile;
		/** Mount name where the temp file is stored */
		OFString m_mount;
		/** Hash directory within the mount temp directory */
		OFString m_hash;
		/** controls whether or not register the object directly */
		RegistrationMode m_registrationMode;
		/** Is double check after moving to repository - used when migrating from classic archive to repository */
		bool m_isDoubleCheck;
		/** Indicates whether registration tasks were started from obsolete tool */
		bool m_isStartedFromObsolete;
		/** true if this is an image (has pixel data) **/
		bool m_hasPixels;
		/** true indicates soft edit values has changed in the PbR object to register */
		bool m_hasEditChanges;
		/** ID of the sender device */
		OFString m_studyDate;

		/** ID of the sender device */
		OFString m_senderDevice;
		/** IP address of the sender device */
		OFString m_senderIP;
		/** AE title of the sender device */
		OFString m_senderAE;

		// Taskd related properties - may be moved to a TaskStarter class

		/** Delta priority */
		OFString m_delta;
		/** Job ID - Identifies the job the task belongs to */
		OFString m_jobId;
		/** Sub Job ID - Identifies the sub job the task belongs to */
		OFString m_subJobId;
		/** Child ID  - Identifies the child task within sub job */
		OFString m_childId;
		/** Job Information */
		OFString m_jobInfo;
		/** Relative Priority */
		int m_relativePriority;

	public:

		/**
		 * Constructor creating a new StorescpRegStarter object
		 *
		 * @param study         Instance UID of the study the object to be stored for
		 * @param file          File name of the object
		 * @param reindexMode   Reindex mode
		 * @param tempFile      Path to the object file to be stored
		 * @param mount         The mount point where the study resides (or will reside) in the DICOM repository
		 */
		StorescpRegStarter( const OFString& study, const OFString& file, STORESCP_REG_OPT reindexMode, const OFString& tempFile,
				const OFString& mount = "" );

		/**
		 * Destructor destroying this StorescpRegStarter object
		 */
		~StorescpRegStarter();

		/**
		 * Sets the ID of the device the object was received from.
		 *
		 * @param senderDevice  ID of the device the object was received from
		 */
		inline void setSenderDevice( const char* senderDevice ) {
			m_senderDevice = senderDevice == NULL ? "" : senderDevice;
		}

		/**
		 * Sets the study date
		 *
		 * @param studyDate  study date from dicom tag (0x0008,0x0020)
		 */
		inline void setStudyDate( const char* studyDate ) {
			m_studyDate = studyDate == NULL ? "" : studyDate;
		}

		/**
		 * Sets the IP address of the device the object was received from.
		 * To be used only in promiscuous mode for unregistered devices,
		 * or in case of objects received from archive
		 *
		 * @param senderIP  IP address of the device the object was received from
		 */
		inline void setSenderIP( const char* senderIP ) {
			m_senderIP = senderIP == NULL ? "" : senderIP;
		}

		/**
		 * Sets the AE title of the device the object was received from.
		 * To be used only in promiscuous mode for unregistered devices.
		 *
		 * @param senderAE  AE title of the device the object was received from
		 */
		inline void setSenderAE( const char* senderAE ) {
			m_senderAE = senderAE == NULL ? "" : senderAE;
		}

		/**
		 * Sets whether the object should be registered directly or not (placing to the queue)
		 * @param registrationMode  controls how the object to be registered
		 */
		inline void setRegistrationMode( RegistrationMode registrationMode ) {
			m_registrationMode = registrationMode;
		}

		/**
		 * Sets whether the object should be double-checked after moving to repository
		 * (migration from classic archive to repository archive)
		 * @param isDoubleCheck  true indicates to double-check object when moving to repository
		 */
		inline void setIsDoubleCheck( bool isDoubleCheck ) {
			m_isDoubleCheck = isDoubleCheck;
		}

		/**
		 * Sets whether the task is started from obsolete tool
		 * ( if true then imagemed/dcprocessing/StudyRemover.postObsolete() Java method should not be invoked )
		 * @param isStartedFromObsolete true indicates the task is started from obsolete tool
		 */
		inline void setIsStartedFromObsolete( bool isStartedFromObsolete ) {
			m_isStartedFromObsolete = isStartedFromObsolete;
		}

		/**
		 * Sets whether the task is to process an actual image object
		 * ( if true, taskd can attempt to batch dcreg processing )
		 * @param hasPixels true indicates the task is to process an actual image object
		 */
		inline void setHasPixels( bool hasPixels ) {
			m_hasPixels = hasPixels;
		}

		/**
		 * Sets Delta priority to be used for creating the task
		 *
		 * @param delta  Delta priority to be used for creating the task
		 */
		inline void setDelta( const char* delta ) {
			m_delta = delta == NULL ? "" : delta;
		}

		/**
		 * Sets the job ID the task to be added to
		 *
		 * @param jobId  ID of the job the task to be added to
		 */
		inline void setJobId( const char* jobId ) {
			m_jobId = jobId == NULL ? "" : jobId;
		}

		/**
		 * Sets the subjob ID the task to be added to
		 *
		 * @param subJobId  ID of the subjob the task to be added to
		 */
		inline void setSubJobId( const char* subJobId ) {
			m_subJobId = subJobId == NULL ? "" : subJobId;
		}

		/**
		 * Sets child ID for the task
		 *
		 * @param childId  child ID to be used when constructing task name
		 */
		inline void setChildId( const char* childId ) {
			m_childId = childId == NULL ? "" : childId;
		}

		/**
		 * Sets job information
		 *
		 * @param jobInfo  job information
		 */
		inline void setJobInfo( const char* jobInfo ) {
			m_jobInfo = jobInfo == NULL ? "" : jobInfo;
		}

		/**
		 * Sets relative priority to be used
		 *
		 * @param relativePriority   Relative priority
		 */
		inline void setRelativePriority( int relativePriority ) {
			m_relativePriority = relativePriority;
		}

		/**
		 * Sets whether soft edit values in the PbR object to register has changed.
		 * @param hasEditChanges new value indicating whether or not soft edit values has changed in the object to register
		 */
		inline void setHasEditChanges( bool hasEditChanges ) {
			m_hasEditChanges = hasEditChanges;
		}

		/**
		 * Sets the hash directory name within the mount temporary directory.
		 * (The first level of the study hash)
		 *
		 * @param hash the hash directory
		 */
		inline void setHash( const std::string& hash ) {
			m_hash = hash;
		}

		/**
		 * Place the object to data/tmp, create and schedule the StorescpReg task
		 *
		 * @return  true if and only if the task was successfully created
		 */
		bool startTask( erad::db::DBConnector& db );
};

/*********************************************************
 *                                                       *
 *               MPPS Forward task helper                *
 *                                                       *
 *********************************************************/
using namespace std;
/// @brief MPPS forward helper class
class MPPSFwrdRegStarter {
	private:
		/** ID of the MPPS request */
		string m_mppsUID;
		/** Referenced Study UID */
		string m_refstudyUID;
		/** Referenced SOP insatnce UID files */
		string m_payLoadfile;
		/** MPPS status - Inprogress / Completed / discontinued*/
		string m_mppsStatus;

	public:
		/**
		 * @brief Parametrised Constructor for MPPSFwrdRegStarter class
		 * @param mppsUID - job information
		 * @param rfstudyUID - study UID
		 * @param payLoadFile - mpps payload file name
		 * @param mppsStatus - mpps status
		 */
		MPPSFwrdRegStarter(const string& mppsUID,
						   const string& rfstudyUID,
						   const string& payLoadFile,
						   const string& mppsStatus );
		/**
		 *  @brief destructor for MPPSFwrdRegStarter class
		 */
		~MPPSFwrdRegStarter();
		/**
		 *  @brief Function to initiate Task
		 *  @return true if success otherwise false
		 */
		bool startTask();
};


#endif /* TASKHELPER_H_ */
