#!/bin/bash
#
# Component: tools
#
# Description:
#   Checks if all the DICOM objects stored in the database are
#   properly archived
#
#
# Usage:
#   checkarchive.sh [-f]
#
#    -f :  Fix archive by sending not or inproperly archived files to the
#          archive
#
# IMPORTANT:
#
#  !! Running this script may take a lot of time because it compares all the files in REPOSITORY_DICOM_DATA
#  !! to its archived 'version'
#  !!
#  !! -f option was never tested
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Locked by        : $Locker$
# Symbolic name    : $Name$
# State            : $State$
#

. $HOME/etc/setenv.sh

LOGGER="/usr/bin/logger"
TAG="checkarchive[$$]"

FIXMODE="false"
[ "$1" = "-f" ] && FIXMODE="true"


Archive="no"

[ -f "$ARCHIVE_VAR/archive.cfg" ] && . $ARCHIVE_VAR/archive.cfg

if [ "$Archive" != "yes" ] ; then
	$LOGGER -p $LOG_FACILITY.info -t "$TAG" "No archive has been set"
	echo "No archive has been set."
	exit 1
fi

$LOGGER -p $LOG_FACILITY.info -t "$TAG" "Checking started [FIX: $FIXMODE]"


Pid=$$
DATE=`date +%Y%m%d%H%M%S`
taskCounter=0

echo "select STYIUID from Dcstudy;" | $MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null | \
while read study ; do

	archdir=`echo "select ARCHDIR from Archive where STYIUID='$study';" |\
			$MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null`
	if [ $? -ne 0 ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "MySQL error in select ARCHDIR"
		echo "MySQL error in select ARCHDIR" 1>&2
		exit 1
	fi

	if [ -z "$archdir" ] ; then
		# Not found in archive table

		echo "$study is not in the Archive table"
		$LOGGER -p $LOG_FACILITY.info -t "$TAG" "$study is not in the Archive table"

		[ "true" != "$FIXMODE" ] && continue

		# add checkarchive reference to STUDYDB to prevent deleting the study
		$UTILS_HOME/bin/refcount -E -t "STUDYDB" -i "$study" -a "checkarchive" >>$LOG/refcount.error 2>&1

		if [ $? -ne 0 ] ; then
			$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Could not add 'checkarchive' reference to study $study"
			echo "Could not add 'checkarchive' reference to study $study" 1>&2
			continue
		fi

		# start archiving
		echo "select FNAME from $OBJECTTBL where STYIUID='$study';" | \
		$MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null | \
		while read file ; do
			# start put the file to the archive
			$LOGGER -p $LOG_FACILITY.info -t "$TAG" "start putting $study/$file to the archive"

			# add archive reference to OBJECTFILE. It will be removed by archive/put.sh
			$UTILS_HOME/bin/refcount -a archive -i "${study}/${file}" -t "OBJECTFILE" >>$LOG/refcount.error 2>&1

			DeltaPriority=""
			[ "${file#PbR.}" != "$file" ] && DeltaPriority="DELTA_ARCHIVE_PUT_PBR"

			taskCounter=$(( $taskCounter + 1 ))
			$TASKD_HOME/createtaskfileapp -p "ARCHIVE_PUT" -t "sh" -d "$DeltaPriority" -a "$taskCounter" -c "$ARCHIVE_HOME/put.sh -s \"$study\" -f \"$file\""
		done

	else
		# Found in archive table, check if all the file was archived correctly
		# start archiving
		echo "select FNAME from $OBJECTTBL where STYIUID='$study';" | \
		$MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null | \
		while read file ; do

			cmp `$REPOSITORYHANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`/$file $ARCHIVE_DIR/$archdir/$study/$file
			if [ $? -ne 0 ] ; then
				echo "$study/$file is not the same in the archive and in the database"
				$LOGGER -p $LOG_FACILITY.info -t "$TAG" "$study/$file is not the same in the archive and in the database"

				[ "true" != "$FIXMODE" ] && continue

				# add archive reference to OBJECTFILE. It will be removed by archive/put.sh
				$UTILS_HOME/bin/refcount -E -a archive -i "${study}/${file}" -t "OBJECTFILE" >>$LOG/refcount.error 2>&1

				if [ $? -ne 0 ] ; then
					echo "Could not add 'archive' reference to file $study_$file" 1>&2
					continue
				fi

				# start put the file to the archive
				$LOGGER -p $LOG_FACILITY.info -t "$TAG" "start putting $study/$file to the archive"

				DeltaPriority=""
				[ "${file#PbR.}" != "$file" ] && DeltaPriority="DELTA_ARCHIVE_PUT_PBR"

				taskCounter=$(( $taskCounter + 1 ))
				$TASKD_HOME/createtaskfileapp -p "ARCHIVE_PUT" -t "sh" -d "$DeltaPriority" -a "$taskCounter" -c "$ARCHIVE_HOME/put.sh -s \"$study\" -f \"$file\""
			fi
		done
	fi

done

$LOGGER -p $LOG_FACILITY.info -t "$TAG" "Checking finished [FIX: $FIXMODE]"
