#!/bin/bash

BaseName=$( basename "$0" )


function usage_() {
	cat <<-EOFUSAGE
	      ${BaseName} [-h] [-H] [-s] [-p <pattern>] [<file>]

	         Generates a csv file from benchmark times collected from the logs.

	         Options:
	                  -h           : prints help
	                  -H           : prints detailed help
	                  -s           : print only summary
	                  -p <pattern> : pre-filter <file> or stdin for this pattern
	                                 note, that if a pattern is provided a filter
	                                 for 'bench' is automatically appended to it
	                  <file>       : the log file, if missing then reads stdin
	EOFUSAGE
}

function longhelp_() {
	cat <<-LONGHELP

	         Reads the log file or the standard input which expected to contain benchmark logs collected
	         for a particular operations.

	         By default produces a comma separated list fields,
	            - the first row contains the bench point labels
	            - the second row contains the count of benchmark entries in the logs
	            - the third row contains the sum of the bench values for the particular bench point
	            - the fourth row contains the average value for the particular bench point
	            - subsequent rows contain the values extracted from the log entries

	         The first column is a serial number (for labeling), and
	         the second column contains the overall value for each log entries.

	         If the '-s' option is provided, then only prints the agregated values, starting with the 
	         aggregated values for each bench points, followed by the overall aggregated values 

	         Usage examples:

	                 grep '<pattern>' ~/var/log/info.log | grep benchtime | ${BaseName}

	                 ${BaseName} -p '<pattern>' -s ~/var/log/info.log

	LONGHELP
}


SummaryOnly="0"
Pattern=""

while getopts ":hHp:s" OPT ; do
	case $OPT in
		h)
			usage_
			exit 0
			;;
		H)
			usage_
			longhelp_
			exit 0
			;;
		p)
			Pattern="$OPTARG.* BENCHTIMES "
			;;
		s)
			SummaryOnly="1"
			;;
		?)
			echo "Unrecognized option: -$OPTARG"
			exit 0
			;;
	esac
done

shift $(( $OPTIND-1 ))


awk -vPattern="${Pattern}" '$0 ~ Pattern {print $NF}' $1 \
| awk -vSummaryOnly="${SummaryOnly}" '


function printSummary() {
	PROCINFO["sorted_in"] = "@val_num_asc";

	for ( n in ndxs ) {
		printf "%-20s %20.2f %5s%d/%d)\n", n, sums[n]/cnts[n], "(", sums[n], cnts[n]
	}

	print ""

	printf "%-20s %20d\n", "Count:", cntTotal
	printf "%-20s %20d\n", "Sum:", sumTotal
	if ( cntTotal == 0 ) {
		printf "%-20s %20s\n", "Average:", "N/A"
	}
	else {
		printf "%-20s %20.2f\n", "Average:", sumTotal/cntTotal
	}
}



BEGIN {
	FS=";";
	sumTotal=0;
	cntTotal=0;
}


{
	sumTotal+=$1;
	cntTotal++;

	totals[cntTotal] = $1;

	for ( i = 2; i <= NF ; i++ ) {
		split($i, a, ":");
		sums[ a[1] ] += a[2];
		cnts[ a[1] ]++;

		if ( ndxs[ a[1] ] == "" ) {
			ndxs[ a[1] ] = i;
		}
		matrix[a[1]][cntTotal] = a[2];
	}
}


END {
	if ( SummaryOnly ) {
		printSummary()
		exit 0
	}

	PROCINFO["sorted_in"] = "@val_num_asc";
	printf "MATRIX,OVERALL";
	for ( n in ndxs ) {
		printf "," n;
	}
	print "";

	printf "Count," cntTotal;
	for ( n in ndxs ) {
		printf "," cnts[n];
	}
	print "";

	printf "Sum," sumTotal;
	for ( n in ndxs ) {
		printf "," sums[n];
	}
	print "";

	if ( cntTotal == 0 ) {
		printf "Avg,"
	}
	else {
		printf "Avg,%.2f", sumTotal/cntTotal;
	}
	for ( n in ndxs ) {
		printf ",%.2f", sums[n] / cnts[n];
	}
	print "";

	for ( i = 1; i <= cntTotal; i++ ) {
		printf i "," totals[ i ];
		for ( n in ndxs ) {
			printf "," matrix[ n ][ i ];
		}
		print "";
	}

}'
