#!/bin/bash
#
# Description:
#    Finds and fixes silly user lists.
#
# Usage:
#   For usage see usage_()
#
# Last Updated     : $Author$
# Update Date Time : $Date$
# RCS File Name    : $RCSfile$
# RCS Revision     : $Revision$
# Symbolic name    : $Name$
# State            : $State$
#

. $HOME/etc/setenv.sh

###############################################################################
# cRefr="60"
# cPageSize="50"
###############################################################################

###############################################################################
# SETTING DEFAULTS
###############################################################################
now=`date +%Y-%m-%d-%h-%s`
UserList="${TMP}/$$.userlist.$$"
ls ${USERS_DATA} -1 | grep -v "^#" > "$UserList"
OutputFile="${TMP}/sillyUserLists.$now"
BackupDir="${BACKUP_VAR}/sillyUserLists.$now"
RefreshTime=60
PageSize=50

###############################################################################
# FUNCTIONS
###############################################################################

#
# Prints usage information
#
# $1 - script name
#
function usage_() {
	cat <<-EOFUSAGE
		Usage: $1 [ -F ] [ -o <output file> ] [ -u <username> ] [ -f <userlist file> ] [ -r <RefreshTime> ] [ -s <PageSize> ] [ -h ] [ -l ]
		   -F                    Also correct the silly user's list, not just identify them
		   -o <file>             Name of output file.  Default is ${TMP}/sillyUserLists.Found.<timestamp>
		   -f <file>             Operate on users listed in file.
		   -h                    Display help
		   -l                    List users on the system
		   -u <user>             Only find/fix lists for user.
		   -r <RefreshTime>      Set the threshold for Page Refresh Time.  Default is: $RefreshTime.
		   -s <PageSize>         Set the threshold for Page Size.  Default is: $PageSize.
	EOFUSAGE

	exit 0
}

#
# Lists users on the system
#
function listUsers_ () {
	echo -e "\n\nUsers on the system:\n"
	ls ${USERS_DATA} -1 | grep -v "^#"
	echo -e "\n\n"
	exit 0
} # End function listUsers


###############################################################################
# ARGUMENT PARSING
###############################################################################
while getopts "Fo:f:hlu:r:s:" OPT ; do
	case $OPT in
		F) FIX="TRUE" ;;
		o) OutputFile="$OPTARG" ;;
		f) UserList="$OPTARG" ;;
		h) usage_ `basename $0` ;;
		l) listUsers_ ;;

		u) User="$OPTARG"
		SPECIFICUSERONLY="TRUE"
		;;

		r) RefreshTime="$OPTARG"
		;;

		s) PageSize="$OPTARG"
		;;

		*) echo "No such option $OPTARG"
		usage_ `basename $0`
		;;

	esac
done


###############################################################################
# MAIN
###############################################################################

echo -e  "\n\tList of Offending Worklist filter files, if any, will be saved to: $OutputFile"
if [ "$FIX" = "TRUE" ] ; then
	echo -e "\n\nCaution:  You are running the script in FIX MODE!  User Lists will be modified.  Copies of the original lists"
	echo -e "will be saved to $BackupDir/ directory in a subdirectory named after the users id, "
	echo -e "i.e. If the user is drhoward copies of the original lists would be saved to\n$BackupDir/drhoward/\n"
	echo -e "Are you sure you want to do this?  [ y / n ]\n"
	read answer
	if [ "$answer" != "y" ] ; then
		echo -e "\nYou answered: $answer EXITING..."
		exit 1
	fi
fi
sleep 10

if [ "$SPECIFICUSERONLY" = "TRUE" ] ; then
	echo "$User" > $UserList
fi


DATE=`date +"%Y%m%d%H%M%S"`
PID=$$

for u in `cat $UserList` ; do
	NUMFILTERS=`ls ${USERS_DATA}/$u/filters/ 2>/dev/null | wc -l`
	if [ $NUMFILTERS -eq 0 ] ; then
		#echo "$u has no filters set"
		continue
	fi

	echo "Current User: $u"
	
	(
	IFS=$'\n'
	for f in `ls -1 ${USERS_DATA}/$u/filters/*.rec 2>/dev/null` ; do
		cRefr=300
		cPageSize=20
		. ${f}

		fixArgs=""
		if [ $cRefr -gt 0 -a $cRefr -lt $RefreshTime ] ; then
			fixArgs=" -e 's;^cRefr=.*$;cRefr=\"300\";'"
		fi
		if [ $cPageSize -gt $PageSize ] ; then
			fixArgs="$fixArgs -e 's;^cPageSize=.*$;cPageSize=\"20\";'"
		fi

		[ -z "$fixArgs" ] && continue
		
		echo "User: $u : Filter: $f : Refresh: $cRefr : PageSize: $cPageSize" >> $OutputFile
		
		[ "$FIX" != "TRUE" ] && continue
		
		mkdir -p $BackupDir/$u/filters/
		cp ${f} $BackupDir/$u/filters/`basename ${f}`.ORIG
		eval sed $fixArgs ${f} > ${f}.new
		mv ${f}.new ${f}

		$TASKD_HOME/createtaskfileapp -p "PROLIFERATE" -t "sh" -c "${DOTCOM_HOME}/proliferate.sh -f \"$f\""
	done
	)
done

